import crypto from 'crypto';
import ReactDOM from 'react-dom';
import React, { Component, PureComponent } from 'react';
import PropTypes from 'prop-types';
import colorContrast from 'color-contrast';

function _classCallCheck(instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
}

function _defineProperties(target, props) {
  for (var i = 0; i < props.length; i++) {
    var descriptor = props[i];
    descriptor.enumerable = descriptor.enumerable || false;
    descriptor.configurable = true;
    if ("value" in descriptor) descriptor.writable = true;
    Object.defineProperty(target, descriptor.key, descriptor);
  }
}

function _createClass(Constructor, protoProps, staticProps) {
  if (protoProps) _defineProperties(Constructor.prototype, protoProps);
  if (staticProps) _defineProperties(Constructor, staticProps);
  return Constructor;
}

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

function _objectSpread(target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i] != null ? arguments[i] : {};
    var ownKeys = Object.keys(source);

    if (typeof Object.getOwnPropertySymbols === 'function') {
      ownKeys = ownKeys.concat(Object.getOwnPropertySymbols(source).filter(function (sym) {
        return Object.getOwnPropertyDescriptor(source, sym).enumerable;
      }));
    }

    ownKeys.forEach(function (key) {
      _defineProperty(target, key, source[key]);
    });
  }

  return target;
}

function _inherits(subClass, superClass) {
  if (typeof superClass !== "function" && superClass !== null) {
    throw new TypeError("Super expression must either be null or a function");
  }

  subClass.prototype = Object.create(superClass && superClass.prototype, {
    constructor: {
      value: subClass,
      writable: true,
      configurable: true
    }
  });
  if (superClass) _setPrototypeOf(subClass, superClass);
}

function _getPrototypeOf(o) {
  _getPrototypeOf = Object.setPrototypeOf ? Object.getPrototypeOf : function _getPrototypeOf(o) {
    return o.__proto__ || Object.getPrototypeOf(o);
  };
  return _getPrototypeOf(o);
}

function _setPrototypeOf(o, p) {
  _setPrototypeOf = Object.setPrototypeOf || function _setPrototypeOf(o, p) {
    o.__proto__ = p;
    return o;
  };

  return _setPrototypeOf(o, p);
}

function _objectWithoutPropertiesLoose(source, excluded) {
  if (source == null) return {};
  var target = {};
  var sourceKeys = Object.keys(source);
  var key, i;

  for (i = 0; i < sourceKeys.length; i++) {
    key = sourceKeys[i];
    if (excluded.indexOf(key) >= 0) continue;
    target[key] = source[key];
  }

  return target;
}

function _objectWithoutProperties(source, excluded) {
  if (source == null) return {};

  var target = _objectWithoutPropertiesLoose(source, excluded);

  var key, i;

  if (Object.getOwnPropertySymbols) {
    var sourceSymbolKeys = Object.getOwnPropertySymbols(source);

    for (i = 0; i < sourceSymbolKeys.length; i++) {
      key = sourceSymbolKeys[i];
      if (excluded.indexOf(key) >= 0) continue;
      if (!Object.prototype.propertyIsEnumerable.call(source, key)) continue;
      target[key] = source[key];
    }
  }

  return target;
}

function _assertThisInitialized(self) {
  if (self === void 0) {
    throw new ReferenceError("this hasn't been initialised - super() hasn't been called");
  }

  return self;
}

function _possibleConstructorReturn(self, call) {
  if (call && (typeof call === "object" || typeof call === "function")) {
    return call;
  }

  return _assertThisInitialized(self);
}

function _toConsumableArray(arr) {
  return _arrayWithoutHoles(arr) || _iterableToArray(arr) || _nonIterableSpread();
}

function _arrayWithoutHoles(arr) {
  if (Array.isArray(arr)) {
    for (var i = 0, arr2 = new Array(arr.length); i < arr.length; i++) arr2[i] = arr[i];

    return arr2;
  }
}

function _iterableToArray(iter) {
  if (Symbol.iterator in Object(iter) || Object.prototype.toString.call(iter) === "[object Arguments]") return Array.from(iter);
}

function _nonIterableSpread() {
  throw new TypeError("Invalid attempt to spread non-iterable instance");
}

var commonjsGlobal = typeof window !== 'undefined' ? window : typeof global !== 'undefined' ? global : typeof self !== 'undefined' ? self : {};

function createCommonjsModule(fn, module) {
	return module = { exports: {} }, fn(module, module.exports), module.exports;
}

var collectionUtils = createCommonjsModule(function (module) {

var utils = module.exports = {};

/**
 * Loops through the collection and calls the callback for each element. if the callback returns truthy, the loop is broken and returns the same value.
 * @public
 * @param {*} collection The collection to loop through. Needs to have a length property set and have indices set from 0 to length - 1.
 * @param {function} callback The callback to be called for each element. The element will be given as a parameter to the callback. If this callback returns truthy, the loop is broken and the same value is returned.
 * @returns {*} The value that a callback has returned (if truthy). Otherwise nothing.
 */
utils.forEach = function(collection, callback) {
    for(var i = 0; i < collection.length; i++) {
        var result = callback(collection[i]);
        if(result) {
            return result;
        }
    }
};
});

var elementUtils = function(options) {
    var getState = options.stateHandler.getState;

    /**
     * Tells if the element has been made detectable and ready to be listened for resize events.
     * @public
     * @param {element} The element to check.
     * @returns {boolean} True or false depending on if the element is detectable or not.
     */
    function isDetectable(element) {
        var state = getState(element);
        return state && !!state.isDetectable;
    }

    /**
     * Marks the element that it has been made detectable and ready to be listened for resize events.
     * @public
     * @param {element} The element to mark.
     */
    function markAsDetectable(element) {
        getState(element).isDetectable = true;
    }

    /**
     * Tells if the element is busy or not.
     * @public
     * @param {element} The element to check.
     * @returns {boolean} True or false depending on if the element is busy or not.
     */
    function isBusy(element) {
        return !!getState(element).busy;
    }

    /**
     * Marks the object is busy and should not be made detectable.
     * @public
     * @param {element} element The element to mark.
     * @param {boolean} busy If the element is busy or not.
     */
    function markBusy(element, busy) {
        getState(element).busy = !!busy;
    }

    return {
        isDetectable: isDetectable,
        markAsDetectable: markAsDetectable,
        isBusy: isBusy,
        markBusy: markBusy
    };
};

var listenerHandler = function(idHandler) {
    var eventListeners = {};

    /**
     * Gets all listeners for the given element.
     * @public
     * @param {element} element The element to get all listeners for.
     * @returns All listeners for the given element.
     */
    function getListeners(element) {
        var id = idHandler.get(element);

        if (id === undefined) {
            return [];
        }

        return eventListeners[id] || [];
    }

    /**
     * Stores the given listener for the given element. Will not actually add the listener to the element.
     * @public
     * @param {element} element The element that should have the listener added.
     * @param {function} listener The callback that the element has added.
     */
    function addListener(element, listener) {
        var id = idHandler.get(element);

        if(!eventListeners[id]) {
            eventListeners[id] = [];
        }

        eventListeners[id].push(listener);
    }

    function removeListener(element, listener) {
        var listeners = getListeners(element);
        for (var i = 0, len = listeners.length; i < len; ++i) {
            if (listeners[i] === listener) {
              listeners.splice(i, 1);
              break;
            }
        }
    }

    function removeAllListeners(element) {
      var listeners = getListeners(element);
      if (!listeners) { return; }
      listeners.length = 0;
    }

    return {
        get: getListeners,
        add: addListener,
        removeListener: removeListener,
        removeAllListeners: removeAllListeners
    };
};

var idGenerator = function() {
    var idCount = 1;

    /**
     * Generates a new unique id in the context.
     * @public
     * @returns {number} A unique id in the context.
     */
    function generate() {
        return idCount++;
    }

    return {
        generate: generate
    };
};

var idHandler = function(options) {
    var idGenerator     = options.idGenerator;
    var getState        = options.stateHandler.getState;

    /**
     * Gets the resize detector id of the element.
     * @public
     * @param {element} element The target element to get the id of.
     * @returns {string|number|null} The id of the element. Null if it has no id.
     */
    function getId(element) {
        var state = getState(element);

        if (state && state.id !== undefined) {
            return state.id;
        }

        return null;
    }

    /**
     * Sets the resize detector id of the element. Requires the element to have a resize detector state initialized.
     * @public
     * @param {element} element The target element to set the id of.
     * @returns {string|number|null} The id of the element.
     */
    function setId(element) {
        var state = getState(element);

        if (!state) {
            throw new Error("setId required the element to have a resize detection state.");
        }

        var id = idGenerator.generate();

        state.id = id;

        return id;
    }

    return {
        get: getId,
        set: setId
    };
};

/* global console: false */

/**
 * Reporter that handles the reporting of logs, warnings and errors.
 * @public
 * @param {boolean} quiet Tells if the reporter should be quiet or not.
 */
var reporter = function(quiet) {
    function noop() {
        //Does nothing.
    }

    var reporter = {
        log: noop,
        warn: noop,
        error: noop
    };

    if(!quiet && window.console) {
        var attachFunction = function(reporter, name) {
            //The proxy is needed to be able to call the method with the console context,
            //since we cannot use bind.
            reporter[name] = function reporterProxy() {
                var f = console[name];
                if (f.apply) { //IE9 does not support console.log.apply :)
                    f.apply(console, arguments);
                } else {
                    for (var i = 0; i < arguments.length; i++) {
                        f(arguments[i]);
                    }
                }
            };
        };

        attachFunction(reporter, "log");
        attachFunction(reporter, "warn");
        attachFunction(reporter, "error");
    }

    return reporter;
};

var browserDetector = createCommonjsModule(function (module) {

var detector = module.exports = {};

detector.isIE = function(version) {
    function isAnyIeVersion() {
        var agent = navigator.userAgent.toLowerCase();
        return agent.indexOf("msie") !== -1 || agent.indexOf("trident") !== -1 || agent.indexOf(" edge/") !== -1;
    }

    if(!isAnyIeVersion()) {
        return false;
    }

    if(!version) {
        return true;
    }

    //Shamelessly stolen from https://gist.github.com/padolsey/527683
    var ieVersion = (function(){
        var undef,
            v = 3,
            div = document.createElement("div"),
            all = div.getElementsByTagName("i");

        do {
            div.innerHTML = "<!--[if gt IE " + (++v) + "]><i></i><![endif]-->";
        }
        while (all[0]);

        return v > 4 ? v : undef;
    }());

    return version === ieVersion;
};

detector.isLegacyOpera = function() {
    return !!window.opera;
};
});

var utils_1 = createCommonjsModule(function (module) {

var utils = module.exports = {};

utils.getOption = getOption;

function getOption(options, name, defaultValue) {
    var value = options[name];

    if((value === undefined || value === null) && defaultValue !== undefined) {
        return defaultValue;
    }

    return value;
}
});

var batchProcessor = function batchProcessorMaker(options) {
    options             = options || {};
    var reporter        = options.reporter;
    var asyncProcess    = utils_1.getOption(options, "async", true);
    var autoProcess     = utils_1.getOption(options, "auto", true);

    if(autoProcess && !asyncProcess) {
        reporter && reporter.warn("Invalid options combination. auto=true and async=false is invalid. Setting async=true.");
        asyncProcess = true;
    }

    var batch = Batch();
    var asyncFrameHandler;
    var isProcessing = false;

    function addFunction(level, fn) {
        if(!isProcessing && autoProcess && asyncProcess && batch.size() === 0) {
            // Since this is async, it is guaranteed to be executed after that the fn is added to the batch.
            // This needs to be done before, since we're checking the size of the batch to be 0.
            processBatchAsync();
        }

        batch.add(level, fn);
    }

    function processBatch() {
        // Save the current batch, and create a new batch so that incoming functions are not added into the currently processing batch.
        // Continue processing until the top-level batch is empty (functions may be added to the new batch while processing, and so on).
        isProcessing = true;
        while (batch.size()) {
            var processingBatch = batch;
            batch = Batch();
            processingBatch.process();
        }
        isProcessing = false;
    }

    function forceProcessBatch(localAsyncProcess) {
        if (isProcessing) {
            return;
        }

        if(localAsyncProcess === undefined) {
            localAsyncProcess = asyncProcess;
        }

        if(asyncFrameHandler) {
            cancelFrame(asyncFrameHandler);
            asyncFrameHandler = null;
        }

        if(localAsyncProcess) {
            processBatchAsync();
        } else {
            processBatch();
        }
    }

    function processBatchAsync() {
        asyncFrameHandler = requestFrame(processBatch);
    }

    function cancelFrame(listener) {
        // var cancel = window.cancelAnimationFrame || window.mozCancelAnimationFrame || window.webkitCancelAnimationFrame || window.clearTimeout;
        var cancel = clearTimeout;
        return cancel(listener);
    }

    function requestFrame(callback) {
        // var raf = window.requestAnimationFrame || window.mozRequestAnimationFrame || window.webkitRequestAnimationFrame || function(fn) { return window.setTimeout(fn, 20); };
        var raf = function(fn) { return setTimeout(fn, 0); };
        return raf(callback);
    }

    return {
        add: addFunction,
        force: forceProcessBatch
    };
};

function Batch() {
    var batch       = {};
    var size        = 0;
    var topLevel    = 0;
    var bottomLevel = 0;

    function add(level, fn) {
        if(!fn) {
            fn = level;
            level = 0;
        }

        if(level > topLevel) {
            topLevel = level;
        } else if(level < bottomLevel) {
            bottomLevel = level;
        }

        if(!batch[level]) {
            batch[level] = [];
        }

        batch[level].push(fn);
        size++;
    }

    function process() {
        for(var level = bottomLevel; level <= topLevel; level++) {
            var fns = batch[level];

            for(var i = 0; i < fns.length; i++) {
                var fn = fns[i];
                fn();
            }
        }
    }

    function getSize() {
        return size;
    }

    return {
        add: add,
        process: process,
        size: getSize
    };
}

var prop = "_erd";

function initState(element) {
    element[prop] = {};
    return getState(element);
}

function getState(element) {
    return element[prop];
}

function cleanState(element) {
    delete element[prop];
}

var stateHandler = {
    initState: initState,
    getState: getState,
    cleanState: cleanState
};

var object = function(options) {
    options             = options || {};
    var reporter        = options.reporter;
    var batchProcessor  = options.batchProcessor;
    var getState        = options.stateHandler.getState;

    if(!reporter) {
        throw new Error("Missing required dependency: reporter.");
    }

    /**
     * Adds a resize event listener to the element.
     * @public
     * @param {element} element The element that should have the listener added.
     * @param {function} listener The listener callback to be called for each resize event of the element. The element will be given as a parameter to the listener callback.
     */
    function addListener(element, listener) {
        if(!getObject(element)) {
            throw new Error("Element is not detectable by this strategy.");
        }

        function listenerProxy() {
            listener(element);
        }

        if(browserDetector.isIE(8)) {
            //IE 8 does not support object, but supports the resize event directly on elements.
            getState(element).object = {
                proxy: listenerProxy
            };
            element.attachEvent("onresize", listenerProxy);
        } else {
            var object = getObject(element);
            object.contentDocument.defaultView.addEventListener("resize", listenerProxy);
        }
    }

    /**
     * Makes an element detectable and ready to be listened for resize events. Will call the callback when the element is ready to be listened for resize changes.
     * @private
     * @param {object} options Optional options object.
     * @param {element} element The element to make detectable
     * @param {function} callback The callback to be called when the element is ready to be listened for resize changes. Will be called with the element as first parameter.
     */
    function makeDetectable(options, element, callback) {
        if (!callback) {
            callback = element;
            element = options;
            options = null;
        }

        options = options || {};
        var debug = options.debug;

        function injectObject(element, callback) {
            var OBJECT_STYLE = "display: block; position: absolute; top: 0; left: 0; width: 100%; height: 100%; border: none; padding: 0; margin: 0; opacity: 0; z-index: -1000; pointer-events: none;";

            //The target element needs to be positioned (everything except static) so the absolute positioned object will be positioned relative to the target element.

            // Position altering may be performed directly or on object load, depending on if style resolution is possible directly or not.
            var positionCheckPerformed = false;

            // The element may not yet be attached to the DOM, and therefore the style object may be empty in some browsers.
            // Since the style object is a reference, it will be updated as soon as the element is attached to the DOM.
            var style = window.getComputedStyle(element);
            var width = element.offsetWidth;
            var height = element.offsetHeight;

            getState(element).startSize = {
                width: width,
                height: height
            };

            function mutateDom() {
                function alterPositionStyles() {
                    if(style.position === "static") {
                        element.style.position = "relative";

                        var removeRelativeStyles = function(reporter, element, style, property) {
                            function getNumericalValue(value) {
                                return value.replace(/[^-\d\.]/g, "");
                            }

                            var value = style[property];

                            if(value !== "auto" && getNumericalValue(value) !== "0") {
                                reporter.warn("An element that is positioned static has style." + property + "=" + value + " which is ignored due to the static positioning. The element will need to be positioned relative, so the style." + property + " will be set to 0. Element: ", element);
                                element.style[property] = 0;
                            }
                        };

                        //Check so that there are no accidental styles that will make the element styled differently now that is is relative.
                        //If there are any, set them to 0 (this should be okay with the user since the style properties did nothing before [since the element was positioned static] anyway).
                        removeRelativeStyles(reporter, element, style, "top");
                        removeRelativeStyles(reporter, element, style, "right");
                        removeRelativeStyles(reporter, element, style, "bottom");
                        removeRelativeStyles(reporter, element, style, "left");
                    }
                }

                function onObjectLoad() {
                    // The object has been loaded, which means that the element now is guaranteed to be attached to the DOM.
                    if (!positionCheckPerformed) {
                        alterPositionStyles();
                    }

                    /*jshint validthis: true */

                    function getDocument(element, callback) {
                        //Opera 12 seem to call the object.onload before the actual document has been created.
                        //So if it is not present, poll it with an timeout until it is present.
                        //TODO: Could maybe be handled better with object.onreadystatechange or similar.
                        if(!element.contentDocument) {
                            setTimeout(function checkForObjectDocument() {
                                getDocument(element, callback);
                            }, 100);

                            return;
                        }

                        callback(element.contentDocument);
                    }

                    //Mutating the object element here seems to fire another load event.
                    //Mutating the inner document of the object element is fine though.
                    var objectElement = this;

                    //Create the style element to be added to the object.
                    getDocument(objectElement, function onObjectDocumentReady(objectDocument) {
                        //Notify that the element is ready to be listened to.
                        callback(element);
                    });
                }

                // The element may be detached from the DOM, and some browsers does not support style resolving of detached elements.
                // The alterPositionStyles needs to be delayed until we know the element has been attached to the DOM (which we are sure of when the onObjectLoad has been fired), if style resolution is not possible.
                if (style.position !== "") {
                    alterPositionStyles(style);
                    positionCheckPerformed = true;
                }

                //Add an object element as a child to the target element that will be listened to for resize events.
                var object = document.createElement("object");
                object.style.cssText = OBJECT_STYLE;
                object.tabIndex = -1;
                object.type = "text/html";
                object.onload = onObjectLoad;

                //Safari: This must occur before adding the object to the DOM.
                //IE: Does not like that this happens before, even if it is also added after.
                if(!browserDetector.isIE()) {
                    object.data = "about:blank";
                }

                element.appendChild(object);
                getState(element).object = object;

                //IE: This must occur after adding the object to the DOM.
                if(browserDetector.isIE()) {
                    object.data = "about:blank";
                }
            }

            if(batchProcessor) {
                batchProcessor.add(mutateDom);
            } else {
                mutateDom();
            }
        }

        if(browserDetector.isIE(8)) {
            //IE 8 does not support objects properly. Luckily they do support the resize event.
            //So do not inject the object and notify that the element is already ready to be listened to.
            //The event handler for the resize event is attached in the utils.addListener instead.
            callback(element);
        } else {
            injectObject(element, callback);
        }
    }

    /**
     * Returns the child object of the target element.
     * @private
     * @param {element} element The target element.
     * @returns The object element of the target.
     */
    function getObject(element) {
        return getState(element).object;
    }

    function uninstall(element) {
        if(browserDetector.isIE(8)) {
            element.detachEvent("onresize", getState(element).object.proxy);
        } else {
            element.removeChild(getObject(element));
        }
        delete getState(element).object;
    }

    return {
        makeDetectable: makeDetectable,
        addListener: addListener,
        uninstall: uninstall
    };
};

var forEach = collectionUtils.forEach;

var scroll = function(options) {
    options             = options || {};
    var reporter        = options.reporter;
    var batchProcessor  = options.batchProcessor;
    var getState        = options.stateHandler.getState;
    var hasState        = options.stateHandler.hasState;
    var idHandler       = options.idHandler;

    if (!batchProcessor) {
        throw new Error("Missing required dependency: batchProcessor");
    }

    if (!reporter) {
        throw new Error("Missing required dependency: reporter.");
    }

    //TODO: Could this perhaps be done at installation time?
    var scrollbarSizes = getScrollbarSizes();

    // Inject the scrollbar styling that prevents them from appearing sometimes in Chrome.
    // The injected container needs to have a class, so that it may be styled with CSS (pseudo elements).
    var styleId = "erd_scroll_detection_scrollbar_style";
    var detectionContainerClass = "erd_scroll_detection_container";
    injectScrollStyle(styleId, detectionContainerClass);

    function getScrollbarSizes() {
        var width = 500;
        var height = 500;

        var child = document.createElement("div");
        child.style.cssText = "position: absolute; width: " + width*2 + "px; height: " + height*2 + "px; visibility: hidden; margin: 0; padding: 0;";

        var container = document.createElement("div");
        container.style.cssText = "position: absolute; width: " + width + "px; height: " + height + "px; overflow: scroll; visibility: none; top: " + -width*3 + "px; left: " + -height*3 + "px; visibility: hidden; margin: 0; padding: 0;";

        container.appendChild(child);

        document.body.insertBefore(container, document.body.firstChild);

        var widthSize = width - container.clientWidth;
        var heightSize = height - container.clientHeight;

        document.body.removeChild(container);

        return {
            width: widthSize,
            height: heightSize
        };
    }

    function injectScrollStyle(styleId, containerClass) {
        function injectStyle(style, method) {
            method = method || function (element) {
                document.head.appendChild(element);
            };

            var styleElement = document.createElement("style");
            styleElement.innerHTML = style;
            styleElement.id = styleId;
            method(styleElement);
            return styleElement;
        }

        if (!document.getElementById(styleId)) {
            var containerAnimationClass = containerClass + "_animation";
            var containerAnimationActiveClass = containerClass + "_animation_active";
            var style = "/* Created by the element-resize-detector library. */\n";
            style += "." + containerClass + " > div::-webkit-scrollbar { display: none; }\n\n";
            style += "." + containerAnimationActiveClass + " { -webkit-animation-duration: 0.1s; animation-duration: 0.1s; -webkit-animation-name: " + containerAnimationClass + "; animation-name: " + containerAnimationClass + "; }\n";
            style += "@-webkit-keyframes " + containerAnimationClass +  " { 0% { opacity: 1; } 50% { opacity: 0; } 100% { opacity: 1; } }\n";
            style += "@keyframes " + containerAnimationClass +          " { 0% { opacity: 1; } 50% { opacity: 0; } 100% { opacity: 1; } }";
            injectStyle(style);
        }
    }

    function addAnimationClass(element) {
        element.className += " " + detectionContainerClass + "_animation_active";
    }

    function addEvent(el, name, cb) {
        if (el.addEventListener) {
            el.addEventListener(name, cb);
        } else if(el.attachEvent) {
            el.attachEvent("on" + name, cb);
        } else {
            return reporter.error("[scroll] Don't know how to add event listeners.");
        }
    }

    function removeEvent(el, name, cb) {
        if (el.removeEventListener) {
            el.removeEventListener(name, cb);
        } else if(el.detachEvent) {
            el.detachEvent("on" + name, cb);
        } else {
            return reporter.error("[scroll] Don't know how to remove event listeners.");
        }
    }

    function getExpandElement(element) {
        return getState(element).container.childNodes[0].childNodes[0].childNodes[0];
    }

    function getShrinkElement(element) {
        return getState(element).container.childNodes[0].childNodes[0].childNodes[1];
    }

    /**
     * Adds a resize event listener to the element.
     * @public
     * @param {element} element The element that should have the listener added.
     * @param {function} listener The listener callback to be called for each resize event of the element. The element will be given as a parameter to the listener callback.
     */
    function addListener(element, listener) {
        var listeners = getState(element).listeners;

        if (!listeners.push) {
            throw new Error("Cannot add listener to an element that is not detectable.");
        }

        getState(element).listeners.push(listener);
    }

    /**
     * Makes an element detectable and ready to be listened for resize events. Will call the callback when the element is ready to be listened for resize changes.
     * @private
     * @param {object} options Optional options object.
     * @param {element} element The element to make detectable
     * @param {function} callback The callback to be called when the element is ready to be listened for resize changes. Will be called with the element as first parameter.
     */
    function makeDetectable(options, element, callback) {
        if (!callback) {
            callback = element;
            element = options;
            options = null;
        }

        options = options || {};

        function debug() {
            if (options.debug) {
                var args = Array.prototype.slice.call(arguments);
                args.unshift(idHandler.get(element), "Scroll: ");
                if (reporter.log.apply) {
                    reporter.log.apply(null, args);
                } else {
                    for (var i = 0; i < args.length; i++) {
                        reporter.log(args[i]);
                    }
                }
            }
        }

        function isDetached(element) {
            function isInDocument(element) {
                return element === element.ownerDocument.body || element.ownerDocument.body.contains(element);
            }

            if (!isInDocument(element)) {
                return true;
            }

            // FireFox returns null style in hidden iframes. See https://github.com/wnr/element-resize-detector/issues/68 and https://bugzilla.mozilla.org/show_bug.cgi?id=795520
            if (window.getComputedStyle(element) === null) {
                return true;
            }

            return false;
        }

        function isUnrendered(element) {
            // Check the absolute positioned container since the top level container is display: inline.
            var container = getState(element).container.childNodes[0];
            var style = window.getComputedStyle(container);
            return !style.width || style.width.indexOf("px") === -1; //Can only compute pixel value when rendered.
        }

        function getStyle() {
            // Some browsers only force layouts when actually reading the style properties of the style object, so make sure that they are all read here,
            // so that the user of the function can be sure that it will perform the layout here, instead of later (important for batching).
            var elementStyle            = window.getComputedStyle(element);
            var style                   = {};
            style.position              = elementStyle.position;
            style.width                 = element.offsetWidth;
            style.height                = element.offsetHeight;
            style.top                   = elementStyle.top;
            style.right                 = elementStyle.right;
            style.bottom                = elementStyle.bottom;
            style.left                  = elementStyle.left;
            style.widthCSS              = elementStyle.width;
            style.heightCSS             = elementStyle.height;
            return style;
        }

        function storeStartSize() {
            var style = getStyle();
            getState(element).startSize = {
                width: style.width,
                height: style.height
            };
            debug("Element start size", getState(element).startSize);
        }

        function initListeners() {
            getState(element).listeners = [];
        }

        function storeStyle() {
            debug("storeStyle invoked.");
            if (!getState(element)) {
                debug("Aborting because element has been uninstalled");
                return;
            }

            var style = getStyle();
            getState(element).style = style;
        }

        function storeCurrentSize(element, width, height) {
            getState(element).lastWidth = width;
            getState(element).lastHeight  = height;
        }

        function getExpandChildElement(element) {
            return getExpandElement(element).childNodes[0];
        }

        function getWidthOffset() {
            return 2 * scrollbarSizes.width + 1;
        }

        function getHeightOffset() {
            return 2 * scrollbarSizes.height + 1;
        }

        function getExpandWidth(width) {
            return width + 10 + getWidthOffset();
        }

        function getExpandHeight(height) {
            return height + 10 + getHeightOffset();
        }

        function getShrinkWidth(width) {
            return width * 2 + getWidthOffset();
        }

        function getShrinkHeight(height) {
            return height * 2 + getHeightOffset();
        }

        function positionScrollbars(element, width, height) {
            var expand          = getExpandElement(element);
            var shrink          = getShrinkElement(element);
            var expandWidth     = getExpandWidth(width);
            var expandHeight    = getExpandHeight(height);
            var shrinkWidth     = getShrinkWidth(width);
            var shrinkHeight    = getShrinkHeight(height);
            expand.scrollLeft   = expandWidth;
            expand.scrollTop    = expandHeight;
            shrink.scrollLeft   = shrinkWidth;
            shrink.scrollTop    = shrinkHeight;
        }

        function injectContainerElement() {
            var container = getState(element).container;

            if (!container) {
                container                   = document.createElement("div");
                container.className         = detectionContainerClass;
                container.style.cssText     = "visibility: hidden; display: inline; width: 0px; height: 0px; z-index: -1; overflow: hidden; margin: 0; padding: 0;";
                getState(element).container = container;
                addAnimationClass(container);
                element.appendChild(container);

                var onAnimationStart = function () {
                    getState(element).onRendered && getState(element).onRendered();
                };

                addEvent(container, "animationstart", onAnimationStart);

                // Store the event handler here so that they may be removed when uninstall is called.
                // See uninstall function for an explanation why it is needed.
                getState(element).onAnimationStart = onAnimationStart;
            }

            return container;
        }

        function injectScrollElements() {
            function alterPositionStyles() {
                var style = getState(element).style;

                if(style.position === "static") {
                    element.style.position = "relative";

                    var removeRelativeStyles = function(reporter, element, style, property) {
                        function getNumericalValue(value) {
                            return value.replace(/[^-\d\.]/g, "");
                        }

                        var value = style[property];

                        if(value !== "auto" && getNumericalValue(value) !== "0") {
                            reporter.warn("An element that is positioned static has style." + property + "=" + value + " which is ignored due to the static positioning. The element will need to be positioned relative, so the style." + property + " will be set to 0. Element: ", element);
                            element.style[property] = 0;
                        }
                    };

                    //Check so that there are no accidental styles that will make the element styled differently now that is is relative.
                    //If there are any, set them to 0 (this should be okay with the user since the style properties did nothing before [since the element was positioned static] anyway).
                    removeRelativeStyles(reporter, element, style, "top");
                    removeRelativeStyles(reporter, element, style, "right");
                    removeRelativeStyles(reporter, element, style, "bottom");
                    removeRelativeStyles(reporter, element, style, "left");
                }
            }

            function getLeftTopBottomRightCssText(left, top, bottom, right) {
                left = (!left ? "0" : (left + "px"));
                top = (!top ? "0" : (top + "px"));
                bottom = (!bottom ? "0" : (bottom + "px"));
                right = (!right ? "0" : (right + "px"));

                return "left: " + left + "; top: " + top + "; right: " + right + "; bottom: " + bottom + ";";
            }

            debug("Injecting elements");

            if (!getState(element)) {
                debug("Aborting because element has been uninstalled");
                return;
            }

            alterPositionStyles();

            var rootContainer = getState(element).container;

            if (!rootContainer) {
                rootContainer = injectContainerElement();
            }

            // Due to this WebKit bug https://bugs.webkit.org/show_bug.cgi?id=80808 (currently fixed in Blink, but still present in WebKit browsers such as Safari),
            // we need to inject two containers, one that is width/height 100% and another that is left/top -1px so that the final container always is 1x1 pixels bigger than
            // the targeted element.
            // When the bug is resolved, "containerContainer" may be removed.

            // The outer container can occasionally be less wide than the targeted when inside inline elements element in WebKit (see https://bugs.webkit.org/show_bug.cgi?id=152980).
            // This should be no problem since the inner container either way makes sure the injected scroll elements are at least 1x1 px.

            var scrollbarWidth          = scrollbarSizes.width;
            var scrollbarHeight         = scrollbarSizes.height;
            var containerContainerStyle = "position: absolute; flex: none; overflow: hidden; z-index: -1; visibility: hidden; width: 100%; height: 100%; left: 0px; top: 0px;";
            var containerStyle          = "position: absolute; flex: none; overflow: hidden; z-index: -1; visibility: hidden; " + getLeftTopBottomRightCssText(-(1 + scrollbarWidth), -(1 + scrollbarHeight), -scrollbarHeight, -scrollbarWidth);
            var expandStyle             = "position: absolute; flex: none; overflow: scroll; z-index: -1; visibility: hidden; width: 100%; height: 100%;";
            var shrinkStyle             = "position: absolute; flex: none; overflow: scroll; z-index: -1; visibility: hidden; width: 100%; height: 100%;";
            var expandChildStyle        = "position: absolute; left: 0; top: 0;";
            var shrinkChildStyle        = "position: absolute; width: 200%; height: 200%;";

            var containerContainer      = document.createElement("div");
            var container               = document.createElement("div");
            var expand                  = document.createElement("div");
            var expandChild             = document.createElement("div");
            var shrink                  = document.createElement("div");
            var shrinkChild             = document.createElement("div");

            // Some browsers choke on the resize system being rtl, so force it to ltr. https://github.com/wnr/element-resize-detector/issues/56
            // However, dir should not be set on the top level container as it alters the dimensions of the target element in some browsers.
            containerContainer.dir              = "ltr";

            containerContainer.style.cssText    = containerContainerStyle;
            containerContainer.className        = detectionContainerClass;
            container.className                 = detectionContainerClass;
            container.style.cssText             = containerStyle;
            expand.style.cssText                = expandStyle;
            expandChild.style.cssText           = expandChildStyle;
            shrink.style.cssText                = shrinkStyle;
            shrinkChild.style.cssText           = shrinkChildStyle;

            expand.appendChild(expandChild);
            shrink.appendChild(shrinkChild);
            container.appendChild(expand);
            container.appendChild(shrink);
            containerContainer.appendChild(container);
            rootContainer.appendChild(containerContainer);

            function onExpandScroll() {
                getState(element).onExpand && getState(element).onExpand();
            }

            function onShrinkScroll() {
                getState(element).onShrink && getState(element).onShrink();
            }

            addEvent(expand, "scroll", onExpandScroll);
            addEvent(shrink, "scroll", onShrinkScroll);

            // Store the event handlers here so that they may be removed when uninstall is called.
            // See uninstall function for an explanation why it is needed.
            getState(element).onExpandScroll = onExpandScroll;
            getState(element).onShrinkScroll = onShrinkScroll;
        }

        function registerListenersAndPositionElements() {
            function updateChildSizes(element, width, height) {
                var expandChild             = getExpandChildElement(element);
                var expandWidth             = getExpandWidth(width);
                var expandHeight            = getExpandHeight(height);
                expandChild.style.width     = expandWidth + "px";
                expandChild.style.height    = expandHeight + "px";
            }

            function updateDetectorElements(done) {
                var width           = element.offsetWidth;
                var height          = element.offsetHeight;

                debug("Storing current size", width, height);

                // Store the size of the element sync here, so that multiple scroll events may be ignored in the event listeners.
                // Otherwise the if-check in handleScroll is useless.
                storeCurrentSize(element, width, height);

                // Since we delay the processing of the batch, there is a risk that uninstall has been called before the batch gets to execute.
                // Since there is no way to cancel the fn executions, we need to add an uninstall guard to all fns of the batch.

                batchProcessor.add(0, function performUpdateChildSizes() {
                    if (!getState(element)) {
                        debug("Aborting because element has been uninstalled");
                        return;
                    }

                    if (!areElementsInjected()) {
                        debug("Aborting because element container has not been initialized");
                        return;
                    }

                    if (options.debug) {
                        var w = element.offsetWidth;
                        var h = element.offsetHeight;

                        if (w !== width || h !== height) {
                            reporter.warn(idHandler.get(element), "Scroll: Size changed before updating detector elements.");
                        }
                    }

                    updateChildSizes(element, width, height);
                });

                batchProcessor.add(1, function updateScrollbars() {
                    if (!getState(element)) {
                        debug("Aborting because element has been uninstalled");
                        return;
                    }

                    if (!areElementsInjected()) {
                        debug("Aborting because element container has not been initialized");
                        return;
                    }

                    positionScrollbars(element, width, height);
                });

                if (done) {
                    batchProcessor.add(2, function () {
                        if (!getState(element)) {
                            debug("Aborting because element has been uninstalled");
                            return;
                        }

                        if (!areElementsInjected()) {
                          debug("Aborting because element container has not been initialized");
                          return;
                        }

                        done();
                    });
                }
            }

            function areElementsInjected() {
                return !!getState(element).container;
            }

            function notifyListenersIfNeeded() {
                function isFirstNotify() {
                    return getState(element).lastNotifiedWidth === undefined;
                }

                debug("notifyListenersIfNeeded invoked");

                var state = getState(element);

                // Don't notify the if the current size is the start size, and this is the first notification.
                if (isFirstNotify() && state.lastWidth === state.startSize.width && state.lastHeight === state.startSize.height) {
                    return debug("Not notifying: Size is the same as the start size, and there has been no notification yet.");
                }

                // Don't notify if the size already has been notified.
                if (state.lastWidth === state.lastNotifiedWidth && state.lastHeight === state.lastNotifiedHeight) {
                    return debug("Not notifying: Size already notified");
                }


                debug("Current size not notified, notifying...");
                state.lastNotifiedWidth = state.lastWidth;
                state.lastNotifiedHeight = state.lastHeight;
                forEach(getState(element).listeners, function (listener) {
                    listener(element);
                });
            }

            function handleRender() {
                debug("startanimation triggered.");

                if (isUnrendered(element)) {
                    debug("Ignoring since element is still unrendered...");
                    return;
                }

                debug("Element rendered.");
                var expand = getExpandElement(element);
                var shrink = getShrinkElement(element);
                if (expand.scrollLeft === 0 || expand.scrollTop === 0 || shrink.scrollLeft === 0 || shrink.scrollTop === 0) {
                    debug("Scrollbars out of sync. Updating detector elements...");
                    updateDetectorElements(notifyListenersIfNeeded);
                }
            }

            function handleScroll() {
                debug("Scroll detected.");

                if (isUnrendered(element)) {
                    // Element is still unrendered. Skip this scroll event.
                    debug("Scroll event fired while unrendered. Ignoring...");
                    return;
                }

                var width = element.offsetWidth;
                var height = element.offsetHeight;

                if (width !== getState(element).lastWidth || height !== getState(element).lastHeight) {
                    debug("Element size changed.");
                    updateDetectorElements(notifyListenersIfNeeded);
                } else {
                    debug("Element size has not changed (" + width + "x" + height + ").");
                }
            }

            debug("registerListenersAndPositionElements invoked.");

            if (!getState(element)) {
                debug("Aborting because element has been uninstalled");
                return;
            }

            getState(element).onRendered = handleRender;
            getState(element).onExpand = handleScroll;
            getState(element).onShrink = handleScroll;

            var style = getState(element).style;
            updateChildSizes(element, style.width, style.height);
        }

        function finalizeDomMutation() {
            debug("finalizeDomMutation invoked.");

            if (!getState(element)) {
                debug("Aborting because element has been uninstalled");
                return;
            }

            var style = getState(element).style;
            storeCurrentSize(element, style.width, style.height);
            positionScrollbars(element, style.width, style.height);
        }

        function ready() {
            callback(element);
        }

        function install() {
            debug("Installing...");
            initListeners();
            storeStartSize();

            batchProcessor.add(0, storeStyle);
            batchProcessor.add(1, injectScrollElements);
            batchProcessor.add(2, registerListenersAndPositionElements);
            batchProcessor.add(3, finalizeDomMutation);
            batchProcessor.add(4, ready);
        }

        debug("Making detectable...");

        if (isDetached(element)) {
            debug("Element is detached");

            injectContainerElement();

            debug("Waiting until element is attached...");

            getState(element).onRendered = function () {
                debug("Element is now attached");
                install();
            };
        } else {
            install();
        }
    }

    function uninstall(element) {
        var state = getState(element);

        if (!state) {
            // Uninstall has been called on a non-erd element.
            return;
        }

        // Uninstall may have been called in the following scenarios:
        // (1) Right between the sync code and async batch (here state.busy = true, but nothing have been registered or injected).
        // (2) In the ready callback of the last level of the batch by another element (here, state.busy = true, but all the stuff has been injected).
        // (3) After the installation process (here, state.busy = false and all the stuff has been injected).
        // So to be on the safe side, let's check for each thing before removing.

        // We need to remove the event listeners, because otherwise the event might fire on an uninstall element which results in an error when trying to get the state of the element.
        state.onExpandScroll && removeEvent(getExpandElement(element), "scroll", state.onExpandScroll);
        state.onShrinkScroll && removeEvent(getShrinkElement(element), "scroll", state.onShrinkScroll);
        state.onAnimationStart && removeEvent(state.container, "animationstart", state.onAnimationStart);

        state.container && element.removeChild(state.container);
    }

    return {
        makeDetectable: makeDetectable,
        addListener: addListener,
        uninstall: uninstall
    };
};

var forEach$1                 = collectionUtils.forEach;









//Detection strategies.



function isCollection(obj) {
    return Array.isArray(obj) || obj.length !== undefined;
}

function toArray(collection) {
    if (!Array.isArray(collection)) {
        var array = [];
        forEach$1(collection, function (obj) {
            array.push(obj);
        });
        return array;
    } else {
        return collection;
    }
}

function isElement(obj) {
    return obj && obj.nodeType === 1;
}

/**
 * @typedef idHandler
 * @type {object}
 * @property {function} get Gets the resize detector id of the element.
 * @property {function} set Generate and sets the resize detector id of the element.
 */

/**
 * @typedef Options
 * @type {object}
 * @property {boolean} callOnAdd    Determines if listeners should be called when they are getting added.
                                    Default is true. If true, the listener is guaranteed to be called when it has been added.
                                    If false, the listener will not be guarenteed to be called when it has been added (does not prevent it from being called).
 * @property {idHandler} idHandler  A custom id handler that is responsible for generating, setting and retrieving id's for elements.
                                    If not provided, a default id handler will be used.
 * @property {reporter} reporter    A custom reporter that handles reporting logs, warnings and errors.
                                    If not provided, a default id handler will be used.
                                    If set to false, then nothing will be reported.
 * @property {boolean} debug        If set to true, the the system will report debug messages as default for the listenTo method.
 */

/**
 * Creates an element resize detector instance.
 * @public
 * @param {Options?} options Optional global options object that will decide how this instance will work.
 */
var elementResizeDetector = function(options) {
    options = options || {};

    //idHandler is currently not an option to the listenTo function, so it should not be added to globalOptions.
    var idHandler$$1;

    if (options.idHandler) {
        // To maintain compatability with idHandler.get(element, readonly), make sure to wrap the given idHandler
        // so that readonly flag always is true when it's used here. This may be removed next major version bump.
        idHandler$$1 = {
            get: function (element) { return options.idHandler.get(element, true); },
            set: options.idHandler.set
        };
    } else {
        var idGenerator$$1 = idGenerator();
        var defaultIdHandler = idHandler({
            idGenerator: idGenerator$$1,
            stateHandler: stateHandler
        });
        idHandler$$1 = defaultIdHandler;
    }

    //reporter is currently not an option to the listenTo function, so it should not be added to globalOptions.
    var reporter$$1 = options.reporter;

    if(!reporter$$1) {
        //If options.reporter is false, then the reporter should be quiet.
        var quiet = reporter$$1 === false;
        reporter$$1 = reporter(quiet);
    }

    //batchProcessor is currently not an option to the listenTo function, so it should not be added to globalOptions.
    var batchProcessor$$1 = getOption(options, "batchProcessor", batchProcessor({ reporter: reporter$$1 }));

    //Options to be used as default for the listenTo function.
    var globalOptions = {};
    globalOptions.callOnAdd     = !!getOption(options, "callOnAdd", true);
    globalOptions.debug         = !!getOption(options, "debug", false);

    var eventListenerHandler    = listenerHandler(idHandler$$1);
    var elementUtils$$1            = elementUtils({
        stateHandler: stateHandler
    });

    //The detection strategy to be used.
    var detectionStrategy;
    var desiredStrategy = getOption(options, "strategy", "object");
    var strategyOptions = {
        reporter: reporter$$1,
        batchProcessor: batchProcessor$$1,
        stateHandler: stateHandler,
        idHandler: idHandler$$1
    };

    if(desiredStrategy === "scroll") {
        if (browserDetector.isLegacyOpera()) {
            reporter$$1.warn("Scroll strategy is not supported on legacy Opera. Changing to object strategy.");
            desiredStrategy = "object";
        } else if (browserDetector.isIE(9)) {
            reporter$$1.warn("Scroll strategy is not supported on IE9. Changing to object strategy.");
            desiredStrategy = "object";
        }
    }

    if(desiredStrategy === "scroll") {
        detectionStrategy = scroll(strategyOptions);
    } else if(desiredStrategy === "object") {
        detectionStrategy = object(strategyOptions);
    } else {
        throw new Error("Invalid strategy name: " + desiredStrategy);
    }

    //Calls can be made to listenTo with elements that are still being installed.
    //Also, same elements can occur in the elements list in the listenTo function.
    //With this map, the ready callbacks can be synchronized between the calls
    //so that the ready callback can always be called when an element is ready - even if
    //it wasn't installed from the function itself.
    var onReadyCallbacks = {};

    /**
     * Makes the given elements resize-detectable and starts listening to resize events on the elements. Calls the event callback for each event for each element.
     * @public
     * @param {Options?} options Optional options object. These options will override the global options. Some options may not be overriden, such as idHandler.
     * @param {element[]|element} elements The given array of elements to detect resize events of. Single element is also valid.
     * @param {function} listener The callback to be executed for each resize event for each element.
     */
    function listenTo(options, elements, listener) {
        function onResizeCallback(element) {
            var listeners = eventListenerHandler.get(element);
            forEach$1(listeners, function callListenerProxy(listener) {
                listener(element);
            });
        }

        function addListener(callOnAdd, element, listener) {
            eventListenerHandler.add(element, listener);

            if(callOnAdd) {
                listener(element);
            }
        }

        //Options object may be omitted.
        if(!listener) {
            listener = elements;
            elements = options;
            options = {};
        }

        if(!elements) {
            throw new Error("At least one element required.");
        }

        if(!listener) {
            throw new Error("Listener required.");
        }

        if (isElement(elements)) {
            // A single element has been passed in.
            elements = [elements];
        } else if (isCollection(elements)) {
            // Convert collection to array for plugins.
            // TODO: May want to check so that all the elements in the collection are valid elements.
            elements = toArray(elements);
        } else {
            return reporter$$1.error("Invalid arguments. Must be a DOM element or a collection of DOM elements.");
        }

        var elementsReady = 0;

        var callOnAdd = getOption(options, "callOnAdd", globalOptions.callOnAdd);
        var onReadyCallback = getOption(options, "onReady", function noop() {});
        var debug = getOption(options, "debug", globalOptions.debug);

        forEach$1(elements, function attachListenerToElement(element) {
            if (!stateHandler.getState(element)) {
                stateHandler.initState(element);
                idHandler$$1.set(element);
            }

            var id = idHandler$$1.get(element);

            debug && reporter$$1.log("Attaching listener to element", id, element);

            if(!elementUtils$$1.isDetectable(element)) {
                debug && reporter$$1.log(id, "Not detectable.");
                if(elementUtils$$1.isBusy(element)) {
                    debug && reporter$$1.log(id, "System busy making it detectable");

                    //The element is being prepared to be detectable. Do not make it detectable.
                    //Just add the listener, because the element will soon be detectable.
                    addListener(callOnAdd, element, listener);
                    onReadyCallbacks[id] = onReadyCallbacks[id] || [];
                    onReadyCallbacks[id].push(function onReady() {
                        elementsReady++;

                        if(elementsReady === elements.length) {
                            onReadyCallback();
                        }
                    });
                    return;
                }

                debug && reporter$$1.log(id, "Making detectable...");
                //The element is not prepared to be detectable, so do prepare it and add a listener to it.
                elementUtils$$1.markBusy(element, true);
                return detectionStrategy.makeDetectable({ debug: debug }, element, function onElementDetectable(element) {
                    debug && reporter$$1.log(id, "onElementDetectable");

                    if (stateHandler.getState(element)) {
                        elementUtils$$1.markAsDetectable(element);
                        elementUtils$$1.markBusy(element, false);
                        detectionStrategy.addListener(element, onResizeCallback);
                        addListener(callOnAdd, element, listener);

                        // Since the element size might have changed since the call to "listenTo", we need to check for this change,
                        // so that a resize event may be emitted.
                        // Having the startSize object is optional (since it does not make sense in some cases such as unrendered elements), so check for its existance before.
                        // Also, check the state existance before since the element may have been uninstalled in the installation process.
                        var state = stateHandler.getState(element);
                        if (state && state.startSize) {
                            var width = element.offsetWidth;
                            var height = element.offsetHeight;
                            if (state.startSize.width !== width || state.startSize.height !== height) {
                                onResizeCallback(element);
                            }
                        }

                        if(onReadyCallbacks[id]) {
                            forEach$1(onReadyCallbacks[id], function(callback) {
                                callback();
                            });
                        }
                    } else {
                        // The element has been unisntalled before being detectable.
                        debug && reporter$$1.log(id, "Element uninstalled before being detectable.");
                    }

                    delete onReadyCallbacks[id];

                    elementsReady++;
                    if(elementsReady === elements.length) {
                        onReadyCallback();
                    }
                });
            }

            debug && reporter$$1.log(id, "Already detecable, adding listener.");

            //The element has been prepared to be detectable and is ready to be listened to.
            addListener(callOnAdd, element, listener);
            elementsReady++;
        });

        if(elementsReady === elements.length) {
            onReadyCallback();
        }
    }

    function uninstall(elements) {
        if(!elements) {
            return reporter$$1.error("At least one element is required.");
        }

        if (isElement(elements)) {
            // A single element has been passed in.
            elements = [elements];
        } else if (isCollection(elements)) {
            // Convert collection to array for plugins.
            // TODO: May want to check so that all the elements in the collection are valid elements.
            elements = toArray(elements);
        } else {
            return reporter$$1.error("Invalid arguments. Must be a DOM element or a collection of DOM elements.");
        }

        forEach$1(elements, function (element) {
            eventListenerHandler.removeAllListeners(element);
            detectionStrategy.uninstall(element);
            stateHandler.cleanState(element);
        });
    }

    return {
        listenTo: listenTo,
        removeListener: eventListenerHandler.removeListener,
        removeAllListeners: eventListenerHandler.removeAllListeners,
        uninstall: uninstall
    };
};

function getOption(options, name, defaultValue) {
    var value = options[name];

    if((value === undefined || value === null) && defaultValue !== undefined) {
        return defaultValue;
    }

    return value;
}

/**
 * Copyright (c) 2013-present, Facebook, Inc.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */

/**
 * Use invariant() to assert state which your program assumes to be true.
 *
 * Provide sprintf-style format (only %s is supported) and arguments
 * to provide information about what broke and what you were
 * expecting.
 *
 * The invariant message will be stripped in production, but the invariant
 * will remain to ensure logic does not differ in production.
 */

var NODE_ENV = process.env.NODE_ENV;

var invariant = function(condition, format, a, b, c, d, e, f) {
  if (NODE_ENV !== 'production') {
    if (format === undefined) {
      throw new Error('invariant requires an error message argument');
    }
  }

  if (!condition) {
    var error;
    if (format === undefined) {
      error = new Error(
        'Minified exception occurred; use the non-minified dev environment ' +
        'for the full error message and additional helpful warnings.'
      );
    } else {
      var args = [a, b, c, d, e, f];
      var argIndex = 0;
      error = new Error(
        format.replace(/%s/g, function() { return args[argIndex++]; })
      );
      error.name = 'Invariant Violation';
    }

    error.framesToPop = 1; // we don't care about invariant's own frame
    throw error;
  }
};

var invariant_1 = invariant;

/**
 * lodash (Custom Build) <https://lodash.com/>
 * Build: `lodash modularize exports="npm" -o ./`
 * Copyright jQuery Foundation and other contributors <https://jquery.org/>
 * Released under MIT license <https://lodash.com/license>
 * Based on Underscore.js 1.8.3 <http://underscorejs.org/LICENSE>
 * Copyright Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 */

/** Used as the `TypeError` message for "Functions" methods. */
var FUNC_ERROR_TEXT = 'Expected a function';

/** Used as references for various `Number` constants. */
var NAN = 0 / 0;

/** `Object#toString` result references. */
var symbolTag = '[object Symbol]';

/** Used to match leading and trailing whitespace. */
var reTrim = /^\s+|\s+$/g;

/** Used to detect bad signed hexadecimal string values. */
var reIsBadHex = /^[-+]0x[0-9a-f]+$/i;

/** Used to detect binary string values. */
var reIsBinary = /^0b[01]+$/i;

/** Used to detect octal string values. */
var reIsOctal = /^0o[0-7]+$/i;

/** Built-in method references without a dependency on `root`. */
var freeParseInt = parseInt;

/** Detect free variable `global` from Node.js. */
var freeGlobal = typeof commonjsGlobal == 'object' && commonjsGlobal && commonjsGlobal.Object === Object && commonjsGlobal;

/** Detect free variable `self`. */
var freeSelf = typeof self == 'object' && self && self.Object === Object && self;

/** Used as a reference to the global object. */
var root = freeGlobal || freeSelf || Function('return this')();

/** Used for built-in method references. */
var objectProto = Object.prototype;

/**
 * Used to resolve the
 * [`toStringTag`](http://ecma-international.org/ecma-262/7.0/#sec-object.prototype.tostring)
 * of values.
 */
var objectToString = objectProto.toString;

/* Built-in method references for those with the same name as other `lodash` methods. */
var nativeMax = Math.max,
    nativeMin = Math.min;

/**
 * Gets the timestamp of the number of milliseconds that have elapsed since
 * the Unix epoch (1 January 1970 00:00:00 UTC).
 *
 * @static
 * @memberOf _
 * @since 2.4.0
 * @category Date
 * @returns {number} Returns the timestamp.
 * @example
 *
 * _.defer(function(stamp) {
 *   console.log(_.now() - stamp);
 * }, _.now());
 * // => Logs the number of milliseconds it took for the deferred invocation.
 */
var now = function() {
  return root.Date.now();
};

/**
 * Creates a debounced function that delays invoking `func` until after `wait`
 * milliseconds have elapsed since the last time the debounced function was
 * invoked. The debounced function comes with a `cancel` method to cancel
 * delayed `func` invocations and a `flush` method to immediately invoke them.
 * Provide `options` to indicate whether `func` should be invoked on the
 * leading and/or trailing edge of the `wait` timeout. The `func` is invoked
 * with the last arguments provided to the debounced function. Subsequent
 * calls to the debounced function return the result of the last `func`
 * invocation.
 *
 * **Note:** If `leading` and `trailing` options are `true`, `func` is
 * invoked on the trailing edge of the timeout only if the debounced function
 * is invoked more than once during the `wait` timeout.
 *
 * If `wait` is `0` and `leading` is `false`, `func` invocation is deferred
 * until to the next tick, similar to `setTimeout` with a timeout of `0`.
 *
 * See [David Corbacho's article](https://css-tricks.com/debouncing-throttling-explained-examples/)
 * for details over the differences between `_.debounce` and `_.throttle`.
 *
 * @static
 * @memberOf _
 * @since 0.1.0
 * @category Function
 * @param {Function} func The function to debounce.
 * @param {number} [wait=0] The number of milliseconds to delay.
 * @param {Object} [options={}] The options object.
 * @param {boolean} [options.leading=false]
 *  Specify invoking on the leading edge of the timeout.
 * @param {number} [options.maxWait]
 *  The maximum time `func` is allowed to be delayed before it's invoked.
 * @param {boolean} [options.trailing=true]
 *  Specify invoking on the trailing edge of the timeout.
 * @returns {Function} Returns the new debounced function.
 * @example
 *
 * // Avoid costly calculations while the window size is in flux.
 * jQuery(window).on('resize', _.debounce(calculateLayout, 150));
 *
 * // Invoke `sendMail` when clicked, debouncing subsequent calls.
 * jQuery(element).on('click', _.debounce(sendMail, 300, {
 *   'leading': true,
 *   'trailing': false
 * }));
 *
 * // Ensure `batchLog` is invoked once after 1 second of debounced calls.
 * var debounced = _.debounce(batchLog, 250, { 'maxWait': 1000 });
 * var source = new EventSource('/stream');
 * jQuery(source).on('message', debounced);
 *
 * // Cancel the trailing debounced invocation.
 * jQuery(window).on('popstate', debounced.cancel);
 */
function debounce(func, wait, options) {
  var lastArgs,
      lastThis,
      maxWait,
      result,
      timerId,
      lastCallTime,
      lastInvokeTime = 0,
      leading = false,
      maxing = false,
      trailing = true;

  if (typeof func != 'function') {
    throw new TypeError(FUNC_ERROR_TEXT);
  }
  wait = toNumber(wait) || 0;
  if (isObject(options)) {
    leading = !!options.leading;
    maxing = 'maxWait' in options;
    maxWait = maxing ? nativeMax(toNumber(options.maxWait) || 0, wait) : maxWait;
    trailing = 'trailing' in options ? !!options.trailing : trailing;
  }

  function invokeFunc(time) {
    var args = lastArgs,
        thisArg = lastThis;

    lastArgs = lastThis = undefined;
    lastInvokeTime = time;
    result = func.apply(thisArg, args);
    return result;
  }

  function leadingEdge(time) {
    // Reset any `maxWait` timer.
    lastInvokeTime = time;
    // Start the timer for the trailing edge.
    timerId = setTimeout(timerExpired, wait);
    // Invoke the leading edge.
    return leading ? invokeFunc(time) : result;
  }

  function remainingWait(time) {
    var timeSinceLastCall = time - lastCallTime,
        timeSinceLastInvoke = time - lastInvokeTime,
        result = wait - timeSinceLastCall;

    return maxing ? nativeMin(result, maxWait - timeSinceLastInvoke) : result;
  }

  function shouldInvoke(time) {
    var timeSinceLastCall = time - lastCallTime,
        timeSinceLastInvoke = time - lastInvokeTime;

    // Either this is the first call, activity has stopped and we're at the
    // trailing edge, the system time has gone backwards and we're treating
    // it as the trailing edge, or we've hit the `maxWait` limit.
    return (lastCallTime === undefined || (timeSinceLastCall >= wait) ||
      (timeSinceLastCall < 0) || (maxing && timeSinceLastInvoke >= maxWait));
  }

  function timerExpired() {
    var time = now();
    if (shouldInvoke(time)) {
      return trailingEdge(time);
    }
    // Restart the timer.
    timerId = setTimeout(timerExpired, remainingWait(time));
  }

  function trailingEdge(time) {
    timerId = undefined;

    // Only invoke if we have `lastArgs` which means `func` has been
    // debounced at least once.
    if (trailing && lastArgs) {
      return invokeFunc(time);
    }
    lastArgs = lastThis = undefined;
    return result;
  }

  function cancel() {
    if (timerId !== undefined) {
      clearTimeout(timerId);
    }
    lastInvokeTime = 0;
    lastArgs = lastCallTime = lastThis = timerId = undefined;
  }

  function flush() {
    return timerId === undefined ? result : trailingEdge(now());
  }

  function debounced() {
    var time = now(),
        isInvoking = shouldInvoke(time);

    lastArgs = arguments;
    lastThis = this;
    lastCallTime = time;

    if (isInvoking) {
      if (timerId === undefined) {
        return leadingEdge(lastCallTime);
      }
      if (maxing) {
        // Handle invocations in a tight loop.
        timerId = setTimeout(timerExpired, wait);
        return invokeFunc(lastCallTime);
      }
    }
    if (timerId === undefined) {
      timerId = setTimeout(timerExpired, wait);
    }
    return result;
  }
  debounced.cancel = cancel;
  debounced.flush = flush;
  return debounced;
}

/**
 * Creates a throttled function that only invokes `func` at most once per
 * every `wait` milliseconds. The throttled function comes with a `cancel`
 * method to cancel delayed `func` invocations and a `flush` method to
 * immediately invoke them. Provide `options` to indicate whether `func`
 * should be invoked on the leading and/or trailing edge of the `wait`
 * timeout. The `func` is invoked with the last arguments provided to the
 * throttled function. Subsequent calls to the throttled function return the
 * result of the last `func` invocation.
 *
 * **Note:** If `leading` and `trailing` options are `true`, `func` is
 * invoked on the trailing edge of the timeout only if the throttled function
 * is invoked more than once during the `wait` timeout.
 *
 * If `wait` is `0` and `leading` is `false`, `func` invocation is deferred
 * until to the next tick, similar to `setTimeout` with a timeout of `0`.
 *
 * See [David Corbacho's article](https://css-tricks.com/debouncing-throttling-explained-examples/)
 * for details over the differences between `_.throttle` and `_.debounce`.
 *
 * @static
 * @memberOf _
 * @since 0.1.0
 * @category Function
 * @param {Function} func The function to throttle.
 * @param {number} [wait=0] The number of milliseconds to throttle invocations to.
 * @param {Object} [options={}] The options object.
 * @param {boolean} [options.leading=true]
 *  Specify invoking on the leading edge of the timeout.
 * @param {boolean} [options.trailing=true]
 *  Specify invoking on the trailing edge of the timeout.
 * @returns {Function} Returns the new throttled function.
 * @example
 *
 * // Avoid excessively updating the position while scrolling.
 * jQuery(window).on('scroll', _.throttle(updatePosition, 100));
 *
 * // Invoke `renewToken` when the click event is fired, but not more than once every 5 minutes.
 * var throttled = _.throttle(renewToken, 300000, { 'trailing': false });
 * jQuery(element).on('click', throttled);
 *
 * // Cancel the trailing throttled invocation.
 * jQuery(window).on('popstate', throttled.cancel);
 */
function throttle(func, wait, options) {
  var leading = true,
      trailing = true;

  if (typeof func != 'function') {
    throw new TypeError(FUNC_ERROR_TEXT);
  }
  if (isObject(options)) {
    leading = 'leading' in options ? !!options.leading : leading;
    trailing = 'trailing' in options ? !!options.trailing : trailing;
  }
  return debounce(func, wait, {
    'leading': leading,
    'maxWait': wait,
    'trailing': trailing
  });
}

/**
 * Checks if `value` is the
 * [language type](http://www.ecma-international.org/ecma-262/7.0/#sec-ecmascript-language-types)
 * of `Object`. (e.g. arrays, functions, objects, regexes, `new Number(0)`, and `new String('')`)
 *
 * @static
 * @memberOf _
 * @since 0.1.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is an object, else `false`.
 * @example
 *
 * _.isObject({});
 * // => true
 *
 * _.isObject([1, 2, 3]);
 * // => true
 *
 * _.isObject(_.noop);
 * // => true
 *
 * _.isObject(null);
 * // => false
 */
function isObject(value) {
  var type = typeof value;
  return !!value && (type == 'object' || type == 'function');
}

/**
 * Checks if `value` is object-like. A value is object-like if it's not `null`
 * and has a `typeof` result of "object".
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is object-like, else `false`.
 * @example
 *
 * _.isObjectLike({});
 * // => true
 *
 * _.isObjectLike([1, 2, 3]);
 * // => true
 *
 * _.isObjectLike(_.noop);
 * // => false
 *
 * _.isObjectLike(null);
 * // => false
 */
function isObjectLike(value) {
  return !!value && typeof value == 'object';
}

/**
 * Checks if `value` is classified as a `Symbol` primitive or object.
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a symbol, else `false`.
 * @example
 *
 * _.isSymbol(Symbol.iterator);
 * // => true
 *
 * _.isSymbol('abc');
 * // => false
 */
function isSymbol(value) {
  return typeof value == 'symbol' ||
    (isObjectLike(value) && objectToString.call(value) == symbolTag);
}

/**
 * Converts `value` to a number.
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to process.
 * @returns {number} Returns the number.
 * @example
 *
 * _.toNumber(3.2);
 * // => 3.2
 *
 * _.toNumber(Number.MIN_VALUE);
 * // => 5e-324
 *
 * _.toNumber(Infinity);
 * // => Infinity
 *
 * _.toNumber('3.2');
 * // => 3.2
 */
function toNumber(value) {
  if (typeof value == 'number') {
    return value;
  }
  if (isSymbol(value)) {
    return NAN;
  }
  if (isObject(value)) {
    var other = typeof value.valueOf == 'function' ? value.valueOf() : value;
    value = isObject(other) ? (other + '') : other;
  }
  if (typeof value != 'string') {
    return value === 0 ? value : +value;
  }
  value = value.replace(reTrim, '');
  var isBinary = reIsBinary.test(value);
  return (isBinary || reIsOctal.test(value))
    ? freeParseInt(value.slice(2), isBinary ? 2 : 8)
    : (reIsBadHex.test(value) ? NAN : +value);
}

var lodash_throttle = throttle;

/**
 * lodash (Custom Build) <https://lodash.com/>
 * Build: `lodash modularize exports="npm" -o ./`
 * Copyright jQuery Foundation and other contributors <https://jquery.org/>
 * Released under MIT license <https://lodash.com/license>
 * Based on Underscore.js 1.8.3 <http://underscorejs.org/LICENSE>
 * Copyright Jeremy Ashkenas, DocumentCloud and Investigative Reporters & Editors
 */

/** Used as the `TypeError` message for "Functions" methods. */
var FUNC_ERROR_TEXT$1 = 'Expected a function';

/** Used as references for various `Number` constants. */
var NAN$1 = 0 / 0;

/** `Object#toString` result references. */
var symbolTag$1 = '[object Symbol]';

/** Used to match leading and trailing whitespace. */
var reTrim$1 = /^\s+|\s+$/g;

/** Used to detect bad signed hexadecimal string values. */
var reIsBadHex$1 = /^[-+]0x[0-9a-f]+$/i;

/** Used to detect binary string values. */
var reIsBinary$1 = /^0b[01]+$/i;

/** Used to detect octal string values. */
var reIsOctal$1 = /^0o[0-7]+$/i;

/** Built-in method references without a dependency on `root`. */
var freeParseInt$1 = parseInt;

/** Detect free variable `global` from Node.js. */
var freeGlobal$1 = typeof commonjsGlobal == 'object' && commonjsGlobal && commonjsGlobal.Object === Object && commonjsGlobal;

/** Detect free variable `self`. */
var freeSelf$1 = typeof self == 'object' && self && self.Object === Object && self;

/** Used as a reference to the global object. */
var root$1 = freeGlobal$1 || freeSelf$1 || Function('return this')();

/** Used for built-in method references. */
var objectProto$1 = Object.prototype;

/**
 * Used to resolve the
 * [`toStringTag`](http://ecma-international.org/ecma-262/7.0/#sec-object.prototype.tostring)
 * of values.
 */
var objectToString$1 = objectProto$1.toString;

/* Built-in method references for those with the same name as other `lodash` methods. */
var nativeMax$1 = Math.max,
    nativeMin$1 = Math.min;

/**
 * Gets the timestamp of the number of milliseconds that have elapsed since
 * the Unix epoch (1 January 1970 00:00:00 UTC).
 *
 * @static
 * @memberOf _
 * @since 2.4.0
 * @category Date
 * @returns {number} Returns the timestamp.
 * @example
 *
 * _.defer(function(stamp) {
 *   console.log(_.now() - stamp);
 * }, _.now());
 * // => Logs the number of milliseconds it took for the deferred invocation.
 */
var now$1 = function() {
  return root$1.Date.now();
};

/**
 * Creates a debounced function that delays invoking `func` until after `wait`
 * milliseconds have elapsed since the last time the debounced function was
 * invoked. The debounced function comes with a `cancel` method to cancel
 * delayed `func` invocations and a `flush` method to immediately invoke them.
 * Provide `options` to indicate whether `func` should be invoked on the
 * leading and/or trailing edge of the `wait` timeout. The `func` is invoked
 * with the last arguments provided to the debounced function. Subsequent
 * calls to the debounced function return the result of the last `func`
 * invocation.
 *
 * **Note:** If `leading` and `trailing` options are `true`, `func` is
 * invoked on the trailing edge of the timeout only if the debounced function
 * is invoked more than once during the `wait` timeout.
 *
 * If `wait` is `0` and `leading` is `false`, `func` invocation is deferred
 * until to the next tick, similar to `setTimeout` with a timeout of `0`.
 *
 * See [David Corbacho's article](https://css-tricks.com/debouncing-throttling-explained-examples/)
 * for details over the differences between `_.debounce` and `_.throttle`.
 *
 * @static
 * @memberOf _
 * @since 0.1.0
 * @category Function
 * @param {Function} func The function to debounce.
 * @param {number} [wait=0] The number of milliseconds to delay.
 * @param {Object} [options={}] The options object.
 * @param {boolean} [options.leading=false]
 *  Specify invoking on the leading edge of the timeout.
 * @param {number} [options.maxWait]
 *  The maximum time `func` is allowed to be delayed before it's invoked.
 * @param {boolean} [options.trailing=true]
 *  Specify invoking on the trailing edge of the timeout.
 * @returns {Function} Returns the new debounced function.
 * @example
 *
 * // Avoid costly calculations while the window size is in flux.
 * jQuery(window).on('resize', _.debounce(calculateLayout, 150));
 *
 * // Invoke `sendMail` when clicked, debouncing subsequent calls.
 * jQuery(element).on('click', _.debounce(sendMail, 300, {
 *   'leading': true,
 *   'trailing': false
 * }));
 *
 * // Ensure `batchLog` is invoked once after 1 second of debounced calls.
 * var debounced = _.debounce(batchLog, 250, { 'maxWait': 1000 });
 * var source = new EventSource('/stream');
 * jQuery(source).on('message', debounced);
 *
 * // Cancel the trailing debounced invocation.
 * jQuery(window).on('popstate', debounced.cancel);
 */
function debounce$1(func, wait, options) {
  var lastArgs,
      lastThis,
      maxWait,
      result,
      timerId,
      lastCallTime,
      lastInvokeTime = 0,
      leading = false,
      maxing = false,
      trailing = true;

  if (typeof func != 'function') {
    throw new TypeError(FUNC_ERROR_TEXT$1);
  }
  wait = toNumber$1(wait) || 0;
  if (isObject$1(options)) {
    leading = !!options.leading;
    maxing = 'maxWait' in options;
    maxWait = maxing ? nativeMax$1(toNumber$1(options.maxWait) || 0, wait) : maxWait;
    trailing = 'trailing' in options ? !!options.trailing : trailing;
  }

  function invokeFunc(time) {
    var args = lastArgs,
        thisArg = lastThis;

    lastArgs = lastThis = undefined;
    lastInvokeTime = time;
    result = func.apply(thisArg, args);
    return result;
  }

  function leadingEdge(time) {
    // Reset any `maxWait` timer.
    lastInvokeTime = time;
    // Start the timer for the trailing edge.
    timerId = setTimeout(timerExpired, wait);
    // Invoke the leading edge.
    return leading ? invokeFunc(time) : result;
  }

  function remainingWait(time) {
    var timeSinceLastCall = time - lastCallTime,
        timeSinceLastInvoke = time - lastInvokeTime,
        result = wait - timeSinceLastCall;

    return maxing ? nativeMin$1(result, maxWait - timeSinceLastInvoke) : result;
  }

  function shouldInvoke(time) {
    var timeSinceLastCall = time - lastCallTime,
        timeSinceLastInvoke = time - lastInvokeTime;

    // Either this is the first call, activity has stopped and we're at the
    // trailing edge, the system time has gone backwards and we're treating
    // it as the trailing edge, or we've hit the `maxWait` limit.
    return (lastCallTime === undefined || (timeSinceLastCall >= wait) ||
      (timeSinceLastCall < 0) || (maxing && timeSinceLastInvoke >= maxWait));
  }

  function timerExpired() {
    var time = now$1();
    if (shouldInvoke(time)) {
      return trailingEdge(time);
    }
    // Restart the timer.
    timerId = setTimeout(timerExpired, remainingWait(time));
  }

  function trailingEdge(time) {
    timerId = undefined;

    // Only invoke if we have `lastArgs` which means `func` has been
    // debounced at least once.
    if (trailing && lastArgs) {
      return invokeFunc(time);
    }
    lastArgs = lastThis = undefined;
    return result;
  }

  function cancel() {
    if (timerId !== undefined) {
      clearTimeout(timerId);
    }
    lastInvokeTime = 0;
    lastArgs = lastCallTime = lastThis = timerId = undefined;
  }

  function flush() {
    return timerId === undefined ? result : trailingEdge(now$1());
  }

  function debounced() {
    var time = now$1(),
        isInvoking = shouldInvoke(time);

    lastArgs = arguments;
    lastThis = this;
    lastCallTime = time;

    if (isInvoking) {
      if (timerId === undefined) {
        return leadingEdge(lastCallTime);
      }
      if (maxing) {
        // Handle invocations in a tight loop.
        timerId = setTimeout(timerExpired, wait);
        return invokeFunc(lastCallTime);
      }
    }
    if (timerId === undefined) {
      timerId = setTimeout(timerExpired, wait);
    }
    return result;
  }
  debounced.cancel = cancel;
  debounced.flush = flush;
  return debounced;
}

/**
 * Checks if `value` is the
 * [language type](http://www.ecma-international.org/ecma-262/7.0/#sec-ecmascript-language-types)
 * of `Object`. (e.g. arrays, functions, objects, regexes, `new Number(0)`, and `new String('')`)
 *
 * @static
 * @memberOf _
 * @since 0.1.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is an object, else `false`.
 * @example
 *
 * _.isObject({});
 * // => true
 *
 * _.isObject([1, 2, 3]);
 * // => true
 *
 * _.isObject(_.noop);
 * // => true
 *
 * _.isObject(null);
 * // => false
 */
function isObject$1(value) {
  var type = typeof value;
  return !!value && (type == 'object' || type == 'function');
}

/**
 * Checks if `value` is object-like. A value is object-like if it's not `null`
 * and has a `typeof` result of "object".
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is object-like, else `false`.
 * @example
 *
 * _.isObjectLike({});
 * // => true
 *
 * _.isObjectLike([1, 2, 3]);
 * // => true
 *
 * _.isObjectLike(_.noop);
 * // => false
 *
 * _.isObjectLike(null);
 * // => false
 */
function isObjectLike$1(value) {
  return !!value && typeof value == 'object';
}

/**
 * Checks if `value` is classified as a `Symbol` primitive or object.
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a symbol, else `false`.
 * @example
 *
 * _.isSymbol(Symbol.iterator);
 * // => true
 *
 * _.isSymbol('abc');
 * // => false
 */
function isSymbol$1(value) {
  return typeof value == 'symbol' ||
    (isObjectLike$1(value) && objectToString$1.call(value) == symbolTag$1);
}

/**
 * Converts `value` to a number.
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to process.
 * @returns {number} Returns the number.
 * @example
 *
 * _.toNumber(3.2);
 * // => 3.2
 *
 * _.toNumber(Number.MIN_VALUE);
 * // => 5e-324
 *
 * _.toNumber(Infinity);
 * // => Infinity
 *
 * _.toNumber('3.2');
 * // => 3.2
 */
function toNumber$1(value) {
  if (typeof value == 'number') {
    return value;
  }
  if (isSymbol$1(value)) {
    return NAN$1;
  }
  if (isObject$1(value)) {
    var other = typeof value.valueOf == 'function' ? value.valueOf() : value;
    value = isObject$1(other) ? (other + '') : other;
  }
  if (typeof value != 'string') {
    return value === 0 ? value : +value;
  }
  value = value.replace(reTrim$1, '');
  var isBinary = reIsBinary$1.test(value);
  return (isBinary || reIsOctal$1.test(value))
    ? freeParseInt$1(value.slice(2), isBinary ? 2 : 8)
    : (reIsBadHex$1.test(value) ? NAN$1 : +value);
}

var lodash_debounce = debounce$1;

//

var shallowequal = function shallowEqual(objA, objB, compare, compareContext) {
  var ret = compare ? compare.call(compareContext, objA, objB) : void 0;

  if (ret !== void 0) {
    return !!ret;
  }

  if (objA === objB) {
    return true;
  }

  if (typeof objA !== "object" || !objA || typeof objB !== "object" || !objB) {
    return false;
  }

  var keysA = Object.keys(objA);
  var keysB = Object.keys(objB);

  if (keysA.length !== keysB.length) {
    return false;
  }

  var bHasOwnProperty = Object.prototype.hasOwnProperty.bind(objB);

  // Test for A's keys different from B.
  for (var idx = 0; idx < keysA.length; idx++) {
    var key = keysA[idx];

    if (!bHasOwnProperty(key)) {
      return false;
    }

    var valueA = objA[key];
    var valueB = objB[key];

    ret = compare ? compare.call(compareContext, valueA, valueB, key) : void 0;

    if (ret === false || (ret === void 0 && valueA !== valueB)) {
      return false;
    }
  }

  return true;
};

function _interopDefault$1 (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var createResizeDetector = _interopDefault$1(elementResizeDetector);

var React__default = _interopDefault$1(React);
var PropTypes$1 = _interopDefault$1(PropTypes);
var ReactDOM$1 = _interopDefault$1(ReactDOM);
var invariant$1 = _interopDefault$1(invariant_1);
var throttle$1 = _interopDefault$1(lodash_throttle);
var debounce$2 = _interopDefault$1(lodash_debounce);
var isShallowEqual = _interopDefault$1(shallowequal);

var instances = {};

// Lazily require to not cause bug
// https://github.com/ctrlplusb/react-sizeme/issues/6
function resizeDetector() {
  var strategy = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : 'scroll';

  if (!instances[strategy]) {
    instances[strategy] = createResizeDetector({
      strategy: strategy
    });
  }

  return instances[strategy];
}

var classCallCheck = function (instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
};

var createClass = function () {
  function defineProperties(target, props) {
    for (var i = 0; i < props.length; i++) {
      var descriptor = props[i];
      descriptor.enumerable = descriptor.enumerable || false;
      descriptor.configurable = true;
      if ("value" in descriptor) descriptor.writable = true;
      Object.defineProperty(target, descriptor.key, descriptor);
    }
  }

  return function (Constructor, protoProps, staticProps) {
    if (protoProps) defineProperties(Constructor.prototype, protoProps);
    if (staticProps) defineProperties(Constructor, staticProps);
    return Constructor;
  };
}();

var _extends$1 = Object.assign || function (target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i];

    for (var key in source) {
      if (Object.prototype.hasOwnProperty.call(source, key)) {
        target[key] = source[key];
      }
    }
  }

  return target;
};

var inherits = function (subClass, superClass) {
  if (typeof superClass !== "function" && superClass !== null) {
    throw new TypeError("Super expression must either be null or a function, not " + typeof superClass);
  }

  subClass.prototype = Object.create(superClass && superClass.prototype, {
    constructor: {
      value: subClass,
      enumerable: false,
      writable: true,
      configurable: true
    }
  });
  if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass;
};

var objectWithoutProperties = function (obj, keys) {
  var target = {};

  for (var i in obj) {
    if (keys.indexOf(i) >= 0) continue;
    if (!Object.prototype.hasOwnProperty.call(obj, i)) continue;
    target[i] = obj[i];
  }

  return target;
};

var possibleConstructorReturn = function (self, call) {
  if (!self) {
    throw new ReferenceError("this hasn't been initialised - super() hasn't been called");
  }

  return call && (typeof call === "object" || typeof call === "function") ? call : self;
};

/* eslint-disable react/no-multi-comp */

var defaultConfig = {
  monitorWidth: true,
  monitorHeight: false,
  monitorPosition: false,
  refreshRate: 16,
  refreshMode: 'throttle',
  noPlaceholder: false,
  resizeDetectorStrategy: 'scroll'
};

function getDisplayName(WrappedComponent) {
  return WrappedComponent.displayName || WrappedComponent.name || 'Component';
}

/**
 * This is a utility wrapper component that will allow our higher order
 * component to get a ref handle on our wrapped components html.
 * @see https://gist.github.com/jimfb/32b587ee6177665fb4cf
 */

var ReferenceWrapper = function (_Component) {
  inherits(ReferenceWrapper, _Component);

  function ReferenceWrapper() {
    classCallCheck(this, ReferenceWrapper);
    return possibleConstructorReturn(this, (ReferenceWrapper.__proto__ || Object.getPrototypeOf(ReferenceWrapper)).apply(this, arguments));
  }

  createClass(ReferenceWrapper, [{
    key: 'render',
    value: function render() {
      return React.Children.only(this.props.children);
    }
  }]);
  return ReferenceWrapper;
}(React.Component);

ReferenceWrapper.displayName = 'SizeMeReferenceWrapper';

ReferenceWrapper.propTypes = { children: PropTypes$1.element.isRequired };

function Placeholder(_ref) {
  var className = _ref.className,
      style = _ref.style;

  // Lets create the props for the temp element.
  var phProps = {};

  // We will use any provided className/style or else make the temp
  // container take the full available space.
  if (!className && !style) {
    phProps.style = { width: '100%', height: '100%' };
  } else {
    if (className) {
      phProps.className = className;
    }
    if (style) {
      phProps.style = style;
    }
  }

  return React__default.createElement('div', phProps);
}
Placeholder.displayName = 'SizeMePlaceholder';
Placeholder.propTypes = {
  className: PropTypes$1.string,
  style: PropTypes$1.object

  /**
   * As we need to maintain a ref on the root node that is rendered within our
   * SizeMe component we need to wrap our entire render in a sub component.
   * Without this, we lose the DOM ref after the placeholder is removed from
   * the render and the actual component is rendered.
   * It took me forever to figure this out, so tread extra careful on this one!
   */
};var renderWrapper = function renderWrapper(WrappedComponent) {
  function SizeMeRenderer(props) {
    var explicitRef = props.explicitRef,
        className = props.className,
        style = props.style,
        size = props.size,
        disablePlaceholder = props.disablePlaceholder,
        onSize = props.onSize,
        restProps = objectWithoutProperties(props, ['explicitRef', 'className', 'style', 'size', 'disablePlaceholder', 'onSize']);


    var noSizeData = size == null || size.width == null && size.height == null && size.position == null;

    var renderPlaceholder = noSizeData && !disablePlaceholder;

    var renderProps = {
      className: className,
      style: style
    };

    if (size != null) {
      renderProps.size = size;
    }

    var toRender = renderPlaceholder ? React__default.createElement(Placeholder, { className: className, style: style }) : React__default.createElement(WrappedComponent, _extends$1({}, renderProps, restProps));

    return React__default.createElement(
      ReferenceWrapper,
      { ref: explicitRef },
      toRender
    );
  }

  SizeMeRenderer.displayName = 'SizeMeRenderer(' + getDisplayName(WrappedComponent) + ')';

  SizeMeRenderer.propTypes = {
    explicitRef: PropTypes$1.func.isRequired,
    className: PropTypes$1.string,
    style: PropTypes$1.object, // eslint-disable-line react/forbid-prop-types
    size: PropTypes$1.shape({
      width: PropTypes$1.number, // eslint-disable-line react/no-unused-prop-types
      height: PropTypes$1.number // eslint-disable-line react/no-unused-prop-types
    }),
    disablePlaceholder: PropTypes$1.bool,
    onSize: PropTypes$1.func
  };

  return SizeMeRenderer;
};

/**
 * :: config -> Component -> WrappedComponent
 *
 * Higher order component that allows the wrapped component to become aware
 * of it's size, by receiving it as an object within it's props.
 *
 * @param  monitorWidth
 *   Default true, whether changes in the element's width should be monitored,
 *   causing a size property to be broadcast.
 * @param  monitorHeight
 *   Default false, whether changes in the element's height should be monitored,
 *   causing a size property to be broadcast.
 *
 * @return The wrapped component.
 */
function withSize() {
  var config = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : defaultConfig;
  var _config$monitorWidth = config.monitorWidth,
      monitorWidth = _config$monitorWidth === undefined ? defaultConfig.monitorWidth : _config$monitorWidth,
      _config$monitorHeight = config.monitorHeight,
      monitorHeight = _config$monitorHeight === undefined ? defaultConfig.monitorHeight : _config$monitorHeight,
      _config$monitorPositi = config.monitorPosition,
      monitorPosition = _config$monitorPositi === undefined ? defaultConfig.monitorPosition : _config$monitorPositi,
      _config$refreshRate = config.refreshRate,
      refreshRate = _config$refreshRate === undefined ? defaultConfig.refreshRate : _config$refreshRate,
      _config$refreshMode = config.refreshMode,
      refreshMode = _config$refreshMode === undefined ? defaultConfig.refreshMode : _config$refreshMode,
      _config$noPlaceholder = config.noPlaceholder,
      noPlaceholder = _config$noPlaceholder === undefined ? defaultConfig.noPlaceholder : _config$noPlaceholder,
      _config$resizeDetecto = config.resizeDetectorStrategy,
      resizeDetectorStrategy = _config$resizeDetecto === undefined ? defaultConfig.resizeDetectorStrategy : _config$resizeDetecto;


  invariant$1(monitorWidth || monitorHeight || monitorPosition, 'You have to monitor at least one of the width, height, or position when using "sizeMe"');

  invariant$1(refreshRate >= 16, "It is highly recommended that you don't put your refreshRate lower than " + '16 as this may cause layout thrashing.');

  invariant$1(refreshMode === 'throttle' || refreshMode === 'debounce', 'The refreshMode should have a value of "throttle" or "debounce"');

  var refreshDelayStrategy = refreshMode === 'throttle' ? throttle$1 : debounce$2;

  return function WrapComponent(WrappedComponent) {
    var SizeMeRenderWrapper = renderWrapper(WrappedComponent);

    var SizeAwareComponent = function (_React$Component) {
      inherits(SizeAwareComponent, _React$Component);

      function SizeAwareComponent() {
        var _ref2;

        var _temp, _this2, _ret;

        classCallCheck(this, SizeAwareComponent);

        for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
          args[_key] = arguments[_key];
        }

        return _ret = (_temp = (_this2 = possibleConstructorReturn(this, (_ref2 = SizeAwareComponent.__proto__ || Object.getPrototypeOf(SizeAwareComponent)).call.apply(_ref2, [this].concat(args))), _this2), _this2.state = {
          width: undefined,
          height: undefined,
          position: undefined
        }, _this2.determineStrategy = function (props) {
          if (props.onSize) {
            if (!_this2.callbackState) {
              _this2.callbackState = _extends$1({}, _this2.state);
            }
            _this2.strategy = 'callback';
          } else {
            _this2.strategy = 'render';
          }
        }, _this2.strategisedSetState = function (state) {
          if (_this2.strategy === 'callback') {
            _this2.callbackState = state;
            _this2.props.onSize(state);
          }
          _this2.setState(state);
        }, _this2.strategisedGetState = function () {
          return _this2.strategy === 'callback' ? _this2.callbackState : _this2.state;
        }, _this2.refCallback = function (element) {
          _this2.element = element;
        }, _this2.hasSizeChanged = function (current, next) {
          var c = current;
          var n = next;
          var cp = c.position || {};
          var np = n.position || {};

          return monitorHeight && c.height !== n.height || monitorPosition && (cp.top !== np.top || cp.left !== np.left || cp.bottom !== np.bottom || cp.right !== np.right) || monitorWidth && c.width !== n.width;
        }, _this2.checkIfSizeChanged = refreshDelayStrategy(function (el) {
          var _el$getBoundingClient = el.getBoundingClientRect(),
              width = _el$getBoundingClient.width,
              height = _el$getBoundingClient.height,
              right = _el$getBoundingClient.right,
              left = _el$getBoundingClient.left,
              top = _el$getBoundingClient.top,
              bottom = _el$getBoundingClient.bottom;

          var next = {
            width: monitorWidth ? width : null,
            height: monitorHeight ? height : null,
            position: monitorPosition ? { right: right, left: left, top: top, bottom: bottom } : null
          };

          if (_this2.hasSizeChanged(_this2.strategisedGetState(), next)) {
            _this2.strategisedSetState(next);
          }
        }, refreshRate), _temp), possibleConstructorReturn(_this2, _ret);
      }

      createClass(SizeAwareComponent, [{
        key: 'componentDidMount',
        value: function componentDidMount() {
          this.detector = resizeDetector(resizeDetectorStrategy);
          this.determineStrategy(this.props);
          this.handleDOMNode(true);
        }
      }, {
        key: 'componentWillReceiveProps',
        value: function componentWillReceiveProps(nextProps) {
          this.determineStrategy(nextProps);
        }
      }, {
        key: 'componentDidUpdate',
        value: function componentDidUpdate() {
          this.handleDOMNode();
        }
      }, {
        key: 'componentWillUnmount',
        value: function componentWillUnmount() {
          // Change our size checker to a noop just in case we have some
          // late running events.
          this.hasSizeChanged = function () {
            return undefined;
          };
          this.checkIfSizeChanged = function () {
            return undefined;
          };

          if (this.domEl) {
            this.detector.uninstall(this.domEl);
            this.domEl = null;
          }
        }
      }, {
        key: 'handleDOMNode',
        value: function handleDOMNode(first) {
          var found = this.element && ReactDOM$1.findDOMNode(this.element);

          if (!found) {
            // If we previously had a dom node then we need to ensure that
            // we remove any existing listeners to avoid memory leaks.
            if (!first && this.domEl) {
              this.detector.removeAllListeners(this.domEl);
              this.domEl = null;
            }
            return;
          }

          if (!first && this.domEl) {
            this.detector.removeAllListeners(this.domEl);
          }

          this.domEl = found;
          this.detector.listenTo(this.domEl, this.checkIfSizeChanged);
        }
      }, {
        key: 'render',
        value: function render() {
          var disablePlaceholder = withSize.enableSSRBehaviour || withSize.noPlaceholders || noPlaceholder || this.strategy === 'callback';

          var size = _extends$1({}, this.state);

          return React__default.createElement(SizeMeRenderWrapper, _extends$1({
            explicitRef: this.refCallback,
            size: this.strategy === 'callback' ? null : size,
            disablePlaceholder: disablePlaceholder
          }, this.props));
        }
      }]);
      return SizeAwareComponent;
    }(React__default.Component);

    SizeAwareComponent.displayName = 'SizeMe(' + getDisplayName(WrappedComponent) + ')';
    SizeAwareComponent.propTypes = {
      onSize: PropTypes$1.func
    };


    SizeAwareComponent.WrappedComponent = WrappedComponent;

    return SizeAwareComponent;
  };
}

/**
 * Allow SizeMe to run within SSR environments.  This is a "global" behaviour
 * flag that should be set within the initialisation phase of your application.
 *
 * Warning: don't set this flag unless you need to as using it may cause
 * extra render cycles to happen within your components depending on the logic
 * contained within them around the usage of the `size` data.
 *
 * DEPRECATED: Please use the global disablePlaceholders
 */
withSize.enableSSRBehaviour = false;

/**
 * Global configuration allowing to disable placeholder rendering for all
 * sizeMe components.
 */
withSize.noPlaceholders = false;

var SizeMe = function (_Component) {
  inherits(SizeMe, _Component);

  function SizeMe(props) {
    classCallCheck(this, SizeMe);

    var _this = possibleConstructorReturn(this, (SizeMe.__proto__ || Object.getPrototypeOf(SizeMe)).call(this, props));

    _initialiseProps.call(_this);

    var children = props.children,
        render = props.render,
        sizeMeConfig = objectWithoutProperties(props, ['children', 'render']);

    _this.createComponent(sizeMeConfig);
    _this.state = {
      size: {
        width: undefined,
        height: undefined
      }
    };
    return _this;
  }

  createClass(SizeMe, [{
    key: 'componentWillReceiveProps',
    value: function componentWillReceiveProps(nextProps) {
      var _props = this.props,
          prevChildren = _props.children,
          prevRender = _props.render,
          prevSizeMeConfig = objectWithoutProperties(_props, ['children', 'render']);
      var nextChildren = nextProps.children,
          nextRender = nextProps.render,
          nextSizeMeConfig = objectWithoutProperties(nextProps, ['children', 'render']);

      if (!isShallowEqual(prevSizeMeConfig, nextSizeMeConfig)) {
        this.createComponent(nextSizeMeConfig);
      }
    }
  }, {
    key: 'render',
    value: function render() {
      var SizeAware = this.SizeAware;

      var render = this.props.children || this.props.render;
      return React__default.createElement(
        SizeAware,
        { onSize: this.onSize },
        render({ size: this.state.size })
      );
    }
  }]);
  return SizeMe;
}(React.Component);

SizeMe.propTypes = {
  children: PropTypes$1.func,
  render: PropTypes$1.func
};
SizeMe.defaultProps = {
  children: undefined,
  render: undefined
};

var _initialiseProps = function _initialiseProps() {
  var _this2 = this;

  this.createComponent = function (config) {
    _this2.SizeAware = withSize(config)(function (_ref) {
      var children = _ref.children;
      return children;
    });
  };

  this.onSize = function (size) {
    return _this2.setState({ size: size });
  };
};

withSize.SizeMe = SizeMe;
withSize.withSize = withSize;

var reactSizeme = withSize;
var reactSizeme_1 = reactSizeme.SizeMe;

var classnames = createCommonjsModule(function (module) {
/*!
  Copyright (c) 2017 Jed Watson.
  Licensed under the MIT License (MIT), see
  http://jedwatson.github.io/classnames
*/
/* global define */

(function () {

	var hasOwn = {}.hasOwnProperty;

	function classNames () {
		var classes = [];

		for (var i = 0; i < arguments.length; i++) {
			var arg = arguments[i];
			if (!arg) continue;

			var argType = typeof arg;

			if (argType === 'string' || argType === 'number') {
				classes.push(arg);
			} else if (Array.isArray(arg) && arg.length) {
				var inner = classNames.apply(null, arg);
				if (inner) {
					classes.push(inner);
				}
			} else if (argType === 'object') {
				for (var key in arg) {
					if (hasOwn.call(arg, key) && arg[key]) {
						classes.push(key);
					}
				}
			}
		}

		return classes.join(' ');
	}

	if (module.exports) {
		classNames.default = classNames;
		module.exports = classNames;
	} else {
		window.classNames = classNames;
	}
}());
});

function styleInject(css, ref) {
  if ( ref === void 0 ) ref = {};
  var insertAt = ref.insertAt;

  if (!css || typeof document === 'undefined') { return; }

  var head = document.head || document.getElementsByTagName('head')[0];
  var style = document.createElement('style');
  style.type = 'text/css';

  if (insertAt === 'top') {
    if (head.firstChild) {
      head.insertBefore(style, head.firstChild);
    } else {
      head.appendChild(style);
    }
  } else {
    head.appendChild(style);
  }

  if (style.styleSheet) {
    style.styleSheet.cssText = css;
  } else {
    style.appendChild(document.createTextNode(css));
  }
}

var css = ".Symbol {\n  height: 100%;\n  width: 100%;\n  display: flex;\n  flex-direction: column;\n  justify-content: flex-end;\n  overflow: hidden;\n  background: none;\n}\n\n.Symbol__image-container {\n  position: relative;\n  flex: 1;\n}\n\n.Symbol__image {\n  position: absolute;\n  top: 0;\n  left: 0;\n  width: 100%;\n  height: 100%;\n  object-fit: contain;\n  pointer-events: none;\n}\n\n.Symbol__label {\n  width: 100%;\n  font-size: 0.875em;\n  font-weight: 700;\n  line-height: 1;\n  text-transform: uppercase;\n  position: relative;\n  text-align: center;\n  display: table-cell;\n  white-space: nowrap;\n  text-overflow: ellipsis;\n  overflow: hidden;\n}\n\n[dir='rtl'] .Symbol__label {\n  direction: rtl;\n}\n\n.Symbol__image-container + .Symbol__label {\n  margin-top: 0.5em;\n}\n";
styleInject(css);

var propTypes = {
  /**
   * Image src to display
   */
  src: PropTypes.string,

  /**
   * Label to display
   */
  label: PropTypes.oneOfType([PropTypes.string, PropTypes.node])
};

function _Symbol(props) {
  var className = props.className,
      label = props.label,
      src = props.src,
      other = _objectWithoutProperties(props, ["className", "label", "src"]);

  var symbolClassName = classnames('Symbol', className);
  return React.createElement("div", _extends({
    className: symbolClassName
  }, other), src && React.createElement("div", {
    className: "Symbol__image-container"
  }, React.createElement("img", {
    className: "Symbol__image",
    src: src,
    alt: "",
    key: src
  })), React.createElement("div", {
    className: "Symbol__label"
  }, label));
}

_Symbol.propTypes = propTypes;

/**
 * Create an empty grid from rows and columns
 *
 * @param {number} rows number of rows
 * @param {number} columns number of columns
 */
function createEmptyGrid(rows, columns) {
  var grid = _toConsumableArray(Array(rows)).map(function () {
    return _toConsumableArray(Array(columns));
  });

  return grid;
}

function fillEmptyCells(grid, items) {
  var itemQueue = _toConsumableArray(items);

  return grid.map(function (row) {
    return row.map(function (item) {
      return item || itemQueue.shift();
    });
  });
}

function iterateGridItems(grid, callback) {
  grid.forEach(function (row, rowIndex) {
    row.forEach(function (item, columnIndex) {
      callback(item, rowIndex, columnIndex);
    });
  });
}

function sortGrid(emptyGrid, order, items) {
  var grid = _toConsumableArray(emptyGrid);

  var itemsToSort = _toConsumableArray(items);

  iterateGridItems(order, function (id, row, column) {
    var itemIndex = itemsToSort.findIndex(function (item) {
      return item.id === id;
    });
    var itemExists = itemIndex > -1;

    if (itemExists) {
      var item = itemsToSort.splice(itemIndex, 1)[0];
      grid[row][column] = item;
    }
  });
  return fillEmptyCells(grid, itemsToSort);
}

var css$1 = ".Row {\r\n  display: flex;\r\n  flex: 1;\r\n  height: var(--row-height);\r\n}\r\n";
styleInject(css$1);

var propTypes$1 = {
  children: PropTypes.node
};

function Row(props) {
  var className = props.className,
      other = _objectWithoutProperties(props, ["className"]);

  var rowClassName = classnames('Row', className);
  return React.createElement("div", _extends({
    className: rowClassName
  }, other));
}

Row.propTypes = propTypes$1;

var css$2 = ".Cell {\n  position: relative;\n  flex: 1;\n  display: flex;\n  justify-content: center;\n  align-items: center;\n  height: calc(100% - (var(--space-unit, 8px) * 2));\n  margin: var(--space-unit, 8px);\n  border-radius: 2px;\n  /* solves rendering issues with iOS */\n  min-width: 0;\n}\n";
styleInject(css$2);

var propTypes$2 = {
  children: PropTypes.node
};

function Cell(props) {
  var className = props.className,
      other = _objectWithoutProperties(props, ["className"]);

  var cellClassName = classnames('Cell', className);
  return React.createElement("div", _extends({
    className: cellClassName
  }, other));
}

Cell.propTypes = propTypes$2;

var css$3 = ".Grid {\n  width: 100%;\n  height: 100%;\n  padding: var(--space-unit, 8px);\n  display: flex;\n  flex-direction: column;\n  overflow: hidden;\n  box-sizing: border-box;\n}\n";
styleInject(css$3);

var isEmptyArray = function isEmptyArray(array) {
  return array.every(function (value) {
    return !value;
  });
};

var Grid =
/*#__PURE__*/
function (_Component) {
  _inherits(Grid, _Component);

  function Grid() {
    var _getPrototypeOf2;

    var _this;

    _classCallCheck(this, Grid);

    for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    _this = _possibleConstructorReturn(this, (_getPrototypeOf2 = _getPrototypeOf(Grid)).call.apply(_getPrototypeOf2, [this].concat(args)));

    _defineProperty(_assertThisInitialized(_assertThisInitialized(_this)), "state", {
      grid: []
    });

    return _this;
  }

  _createClass(Grid, [{
    key: "shouldComponentUpdate",
    value: function shouldComponentUpdate(nextProps) {
      // TODO: optimize when due
      return true;
    }
  }, {
    key: "render",
    value: function render() {
      var _this$props = this.props,
          className = _this$props.className,
          gap = _this$props.gap,
          renderItem = _this$props.renderItem,
          rows = _this$props.rows,
          RowWrapper = _this$props.rowWrapper,
          size = _this$props.size;
      var space = gap / 2;
      var rowHeight = size.height / rows;
      var gridStyle = {
        '--space-unit': "".concat(space, "px"),
        '--row-height': "".concat(rowHeight, "px")
      };
      var gridClassName = classnames('Grid', className);
      return React.createElement("div", {
        className: gridClassName,
        style: gridStyle
      }, this.state.grid.map(function (row, rowIndex) {
        return React.createElement(RowWrapper, {
          key: rowIndex,
          disabled: isEmptyArray(row)
        }, React.createElement(Row, null, row.map(function (item, cellIndex) {
          return React.createElement(Cell, {
            key: cellIndex
          }, item && renderItem(item));
        })));
      }));
    }
  }], [{
    key: "getDerivedStateFromProps",
    value: function getDerivedStateFromProps(nextProps, prevState) {
      var items = nextProps.items,
          columns = nextProps.columns,
          order = nextProps.order,
          rows = nextProps.rows;
      var emptyGrid = createEmptyGrid(rows, columns);
      var sortedGrid = sortGrid(emptyGrid, order, items);
      return {
        grid: sortedGrid
      };
    }
  }]);

  return Grid;
}(Component);

_defineProperty(Grid, "propTypes", {
  /**
   * Items to render
   */
  items: PropTypes.arrayOf(PropTypes.shape({
    id: PropTypes.oneOfType([PropTypes.string, PropTypes.number]).isRequired
  })),

  /**
   * Number of columns
   */
  columns: PropTypes.number,

  /**
   * Gap between cells
   */
  gap: PropTypes.number,

  /**
   * Item order
   */
  order: PropTypes.arrayOf(PropTypes.arrayOf(PropTypes.oneOfType([PropTypes.string, PropTypes.number]))),

  /**
   * Item renderer
   *
   */
  renderItem: PropTypes.func.isRequired,

  /**
   * Number of rows
   */
  rows: PropTypes.number,

  /**
   * Grid size
   */
  size: PropTypes.shape({
    width: PropTypes.number,
    height: PropTypes.number
  })
});

_defineProperty(Grid, "defaultProps", {
  items: [],
  columns: 3,
  gap: 16,
  order: [],
  rows: 3,
  rowWrapper: function rowWrapper(props) {
    return props.children;
  }
});

var BackspaceSvg = (function (props) {
  return React.createElement("svg", _extends({
    xmlns: "http://www.w3.org/2000/svg",
    width: "24",
    height: "24",
    viewBox: "0 0 24 24"
  }, props), React.createElement("path", {
    d: "M0 0h24v24H0z",
    fill: "none"
  }), React.createElement("path", {
    d: "M22 3H7c-.69 0-1.23.35-1.59.88L0 12l5.41 8.11c.36.53.9.89 1.59.89h15c1.1 0 2-.9 2-2V5c0-1.1-.9-2-2-2zm-3 12.59L17.59 17 14 13.41 10.41 17 9 15.59 12.59 12 9 8.41 10.41 7 14 10.59 17.59 7 19 8.41 15.41 12 19 15.59z"
  }));
});

var css$4 = "/* TODO: duplicate code with ClearButton */\n.BackspaceButton {\n  padding: 0.5em;\n  line-height: 0;\n  transition: background-color 0.25s;\n  border-radius: 4px;\n  color: #444;\n}\n\n.BackspaceButton:hover,\n.BackspaceButton:focus {\n  background: rgba(0, 0, 0, 0.15);\n}\n";
styleInject(css$4);

var BackspaceButton =
/*#__PURE__*/
function (_PureComponent) {
  _inherits(BackspaceButton, _PureComponent);

  function BackspaceButton() {
    _classCallCheck(this, BackspaceButton);

    return _possibleConstructorReturn(this, _getPrototypeOf(BackspaceButton).apply(this, arguments));
  }

  _createClass(BackspaceButton, [{
    key: "render",
    value: function render() {
      var _this$props = this.props,
          classNameProp = _this$props.className,
          other = _objectWithoutProperties(_this$props, ["className"]);

      var className = classnames('BackspaceButton', classNameProp);
      return React.createElement("button", _extends({
        className: className,
        "aria-label": "Backspace",
        title: "Backspace"
      }, other), React.createElement(BackspaceSvg, {
        width: "3em",
        height: "3em"
      }));
    }
  }]);

  return BackspaceButton;
}(PureComponent);

var ClearSvg = (function (props) {
  return React.createElement("svg", _extends({
    xmlns: "http://www.w3.org/2000/svg",
    width: "24",
    height: "24",
    viewBox: "0 0 24 24"
  }, props), React.createElement("path", {
    d: "M19 6.41L17.59 5 12 10.59 6.41 5 5 6.41 10.59 12 5 17.59 6.41 19 12 13.41 17.59 19 19 17.59 13.41 12z"
  }), React.createElement("path", {
    d: "M0 0h24v24H0z",
    fill: "none"
  }));
});

var css$5 = "/* TODO: duplicate code with BackspaceButton */\n.ClearButton {\n  padding: 0.5em;\n  line-height: 0;\n  transition: background-color 0.25s;\n  border-radius: 4px;\n  color: #444;\n}\n\n.ClearButton:hover,\n.ClearButton:focus {\n  background: rgba(0, 0, 0, 0.15);\n}\n";
styleInject(css$5);

var ClearButton =
/*#__PURE__*/
function (_PureComponent) {
  _inherits(ClearButton, _PureComponent);

  function ClearButton() {
    _classCallCheck(this, ClearButton);

    return _possibleConstructorReturn(this, _getPrototypeOf(ClearButton).apply(this, arguments));
  }

  _createClass(ClearButton, [{
    key: "render",
    value: function render() {
      var _this$props = this.props,
          className = _this$props.className,
          other = _objectWithoutProperties(_this$props, ["className"]);

      var clearButtonClassName = classnames('ClearButton', className);
      return React.createElement("button", _extends({
        className: clearButtonClassName,
        "aria-label": "Clear",
        title: "Clear"
      }, other), React.createElement(ClearSvg, {
        width: "3em",
        height: "3em"
      }));
    }
  }]);

  return ClearButton;
}(PureComponent);

var css$6 = ".Scroll {\n  position: relative;\n  width: 100%;\n  height: 100%;\n  overflow-x: auto;\n  overflow-y: hidden;\n  outline: 0;\n  -webkit-overflow-scrolling: touch;\n}\n\n.Scroll__container {\n  position: absolute;\n  min-width: 100%;\n  height: 100%;\n  display: flex;\n  align-items: center;\n  outline: 0;\n  transition: background-color 0.25s;\n}\n\n.Scroll__container:not(:empty):focus,\n.Scroll__container:not(:empty):hover {\n  background: rgba(0, 0, 0, 0.15);\n}\n\n.Scroll__container:not(:empty) {\n  cursor: pointer;\n}\n";
styleInject(css$6);

var invertDir = function invertDir(dir) {
  return dir === 'rtl' ? 'ltr' : 'rtl';
};

var Scroll =
/*#__PURE__*/
function (_PureComponent) {
  _inherits(Scroll, _PureComponent);

  function Scroll(props) {
    var _this;

    _classCallCheck(this, Scroll);

    _this = _possibleConstructorReturn(this, _getPrototypeOf(Scroll).call(this, props));
    _this.scrollRef = React.createRef();
    return _this;
  }

  _createClass(Scroll, [{
    key: "scrollToEnd",
    value: function scrollToEnd() {
      var dir = this.props.dir;
      var scrollRef = this.scrollRef.current;

      if (scrollRef.scroll) {
        var scrollXEnd = dir === 'ltr' ? scrollRef.scrollWidth - scrollRef.offsetWidth : 0;
        scrollRef.scroll(scrollXEnd, 0);
      }
    }
  }, {
    key: "componentDidUpdate",
    value: function componentDidUpdate(prevProps) {
      if (this.props.children.length !== prevProps.children.length) {
        this.scrollToEnd();
      }
    }
  }, {
    key: "render",
    value: function render() {
      var _this$props = this.props,
          children = _this$props.children,
          className = _this$props.className,
          dir = _this$props.dir,
          style = _this$props.style,
          tabIndex = _this$props.tabIndex,
          other = _objectWithoutProperties(_this$props, ["children", "className", "dir", "style", "tabIndex"]);

      var scrollClassName = classnames('Scroll', className);
      var scrollStyle = {
        direction: invertDir(dir)
      };
      return React.createElement("div", _extends({
        className: scrollClassName,
        style: scrollStyle
      }, other, {
        ref: this.scrollRef
      }), React.createElement("div", {
        className: "Scroll__container",
        style: _objectSpread({}, style, {
          direction: dir
        }),
        tabIndex: tabIndex
      }, children));
    }
  }]);

  return Scroll;
}(PureComponent);

_defineProperty(Scroll, "propTypes", {
  /**
   * Text direction
   */
  dir: PropTypes.oneOf(['ltr', 'rtl'])
});

_defineProperty(Scroll, "defaultProps", {
  dir: 'ltr'
});

var css$7 = ".Output {\n  height: 100%;\n  width: 100%;\n  display: flex;\n  background: #fff;\n}\n\n.Output__value {\n  box-sizing: border-box;\n  height: 100%;\n  min-width: 7em;\n  padding: 0.5em;\n}\n\n.Output__button {\n  display: flex;\n  flex-direction: column;\n  justify-content: center;\n  margin: 0 0.5em;\n}\n";
styleInject(css$7);

var KeyCodes = {
  enter: 13,
  space: 32
};

var Output =
/*#__PURE__*/
function (_PureComponent) {
  _inherits(Output, _PureComponent);

  function Output() {
    var _getPrototypeOf2;

    var _this;

    _classCallCheck(this, Output);

    for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    _this = _possibleConstructorReturn(this, (_getPrototypeOf2 = _getPrototypeOf(Output)).call.apply(_getPrototypeOf2, [this].concat(args)));

    _defineProperty(_assertThisInitialized(_assertThisInitialized(_this)), "handleBackspace", function (event) {
      _this.outputPop();
    });

    _defineProperty(_assertThisInitialized(_assertThisInitialized(_this)), "handleClear", function (event) {
      _this.outputClear();
    });

    _defineProperty(_assertThisInitialized(_assertThisInitialized(_this)), "handleKeyDown", function (event) {
      if (event.keyCode === KeyCodes.space) {
        event.preventDefault();
      }
    });

    _defineProperty(_assertThisInitialized(_assertThisInitialized(_this)), "handleKeyUp", function (event) {
      var onClick = _this.props.onClick;

      switch (event.keyCode) {
        // fall through
        case KeyCodes.enter:
        case KeyCodes.space:
          onClick && onClick();
          break;

        default: // no default

      }
    });

    return _this;
  }

  _createClass(Output, [{
    key: "outputClear",
    value: function outputClear() {
      var onChange = this.props.onChange;
      var output = [];
      onChange(output);
    }
  }, {
    key: "outputPop",
    value: function outputPop() {
      var _this$props = this.props,
          onChange = _this$props.onChange,
          values = _this$props.values;

      var output = _toConsumableArray(values);

      output.pop();
      onChange(output);
    }
  }, {
    key: "renderClearButton",
    value: function renderClearButton() {
      var _this$props2 = this.props,
          clearButton = _this$props2.clearButton,
          values = _this$props2.values;
      var disabled = !values.length;
      var style = {
        visibility: disabled ? 'hidden' : 'visible'
      };
      var props = {
        disabled: disabled,
        style: style,
        onClick: this.handleClear
      };
      return clearButton ? React.cloneElement(clearButton, props) : React.createElement(ClearButton, props);
    }
  }, {
    key: "renderBackspaceButton",
    value: function renderBackspaceButton() {
      var backspaceButton = this.props.backspaceButton;
      var props = {
        onClick: this.handleBackspace
      };
      return backspaceButton ? React.cloneElement(backspaceButton, props) : React.createElement(BackspaceButton, props);
    }
  }, {
    key: "render",
    value: function render() {
      var _this$props3 = this.props,
          className = _this$props3.className,
          dir = _this$props3.dir,
          InteractiveWrapper = _this$props3.interactiveWrapper,
          onClick = _this$props3.onClick,
          renderValue = _this$props3.renderValue,
          values = _this$props3.values;
      var outputClassName = classnames('Output', className);
      var tabIndex = values.length ? '0' : '-1';
      return React.createElement("div", {
        className: outputClassName
      }, React.createElement(InteractiveWrapper, null, React.createElement(Scroll, {
        dir: dir,
        onClick: onClick,
        onKeyDown: this.handleKeyDown,
        onKeyUp: this.handleKeyUp,
        tabIndex: tabIndex
      }, values.map(function (value, index) {
        return React.createElement("div", {
          className: "Output__value",
          key: index
        }, renderValue(value));
      }))), React.createElement("div", {
        className: "Output__button"
      }, React.createElement(InteractiveWrapper, null, this.renderClearButton())), React.createElement("div", {
        className: "Output__button"
      }, React.createElement(InteractiveWrapper, null, this.renderBackspaceButton())));
    }
  }]);

  return Output;
}(PureComponent);

_defineProperty(Output, "propTypes", {
  /**
   * Backspace button component
   */
  backspaceButton: PropTypes.node,

  /**
   * Clear button component
   */
  clearButton: PropTypes.node,

  /**
   * Text direction
   */
  dir: PropTypes.oneOf(['ltr', 'rtl']),

  /**
   * Callback, fired when output changes
   */
  onChange: PropTypes.func.isRequired,

  /**
   * Callback, fired when output is clicked
   */
  onClick: PropTypes.func,

  /**
   * Render value
   */
  renderValue: PropTypes.func.isRequired,

  /**
   *
   */
  interactiveWrapper: PropTypes.func,

  /**
   * Value to output
   */
  values: PropTypes.array
});

_defineProperty(Output, "defaultProps", {
  interactiveWrapper: function interactiveWrapper(props) {
    return props.children;
  },
  values: []
});

var ButtonActions = {
  backspace: ':backspace',
  clear: ':clear',
  home: ':home',
  space: ':space',
  speak: ':speak'
};

var css$8 = ".Board {\n  width: 100%;\n  height: 100%;\n  display: flex;\n  flex-direction: column;\n  user-select: none;\n}\n\n.Board__output {\n  height: 7.5em;\n}\n\n.Board__main {\n  flex: 1;\n  display: flex;\n  flex-direction: column;\n}\n\n.Board__navbar {\n  flex-shrink: 0;\n}\n\n.Board__name {\n  height: 3em;\n  display: flex;\n  align-items: center;\n  justify-content: center;\n  background: #444;\n  color: #fff;\n}\n\n.Board__buttons {\n  flex: 1;\n  overflow: auto;\n}\n";
styleInject(css$8);

var FontSizes = {
  default: '16px',
  large: '20px',
  larger: '24px'
};
/**
 * BoardLayout component, renders OBF compatible boards
 */

var BoardLayout =
/*#__PURE__*/
function (_PureComponent) {
  _inherits(BoardLayout, _PureComponent);

  function BoardLayout() {
    _classCallCheck(this, BoardLayout);

    return _possibleConstructorReturn(this, _getPrototypeOf(BoardLayout).apply(this, arguments));
  }

  _createClass(BoardLayout, [{
    key: "getFontSize",
    value: function getFontSize(size) {
      return FontSizes[size] || FontSizes.default;
    }
  }, {
    key: "render",
    value: function render() {
      var _this$props = this.props,
          dir = _this$props.dir,
          grid = _this$props.grid,
          navbar = _this$props.navbar,
          output = _this$props.output,
          size = _this$props.size,
          toolbar = _this$props.toolbar;
      var boardStyle = {
        fontSize: this.getFontSize(size)
      };
      return React.createElement("div", {
        className: "Board",
        dir: dir,
        style: boardStyle
      }, output && React.createElement("div", {
        className: "Board__output"
      }, output), React.createElement("div", {
        className: "Board__main"
      }, React.createElement("div", {
        className: "Board__navbar"
      }, navbar), toolbar && React.createElement("div", {
        className: "Board__toolbar"
      }, toolbar), React.createElement("div", {
        className: "Board__buttons"
      }, grid)));
    }
  }]);

  return BoardLayout;
}(PureComponent);

_defineProperty(BoardLayout, "propTypes", {
  /**
   * Text direction
   */
  dir: PropTypes.oneOf(['ltr', 'rtl']),
  grid: PropTypes.node,
  navbar: PropTypes.node,
  output: PropTypes.node,
  size: PropTypes.oneOf(['default', 'large', 'larger']),
  toolbar: PropTypes.node
});

// Unique ID creation requires a high quality random # generator.  In node.js
// this is pretty straight-forward - we use the crypto API.



var rng = function nodeRNG() {
  return crypto.randomBytes(16);
};

/**
 * Convert array of 16 byte values to UUID string format of the form:
 * XXXXXXXX-XXXX-XXXX-XXXX-XXXXXXXXXXXX
 */
var byteToHex = [];
for (var i = 0; i < 256; ++i) {
  byteToHex[i] = (i + 0x100).toString(16).substr(1);
}

function bytesToUuid(buf, offset) {
  var i = offset || 0;
  var bth = byteToHex;
  // join used to fix memory issue caused by concatenation: https://bugs.chromium.org/p/v8/issues/detail?id=3175#c4
  return ([bth[buf[i++]], bth[buf[i++]], 
	bth[buf[i++]], bth[buf[i++]], '-',
	bth[buf[i++]], bth[buf[i++]], '-',
	bth[buf[i++]], bth[buf[i++]], '-',
	bth[buf[i++]], bth[buf[i++]], '-',
	bth[buf[i++]], bth[buf[i++]],
	bth[buf[i++]], bth[buf[i++]],
	bth[buf[i++]], bth[buf[i++]]]).join('');
}

var bytesToUuid_1 = bytesToUuid;

function v4(options, buf, offset) {
  var i = buf && offset || 0;

  if (typeof(options) == 'string') {
    buf = options === 'binary' ? new Array(16) : null;
    options = null;
  }
  options = options || {};

  var rnds = options.random || (options.rng || rng)();

  // Per 4.4, set bits for version and `clock_seq_hi_and_reserved`
  rnds[6] = (rnds[6] & 0x0f) | 0x40;
  rnds[8] = (rnds[8] & 0x3f) | 0x80;

  // Copy bytes to buffer, if provided
  if (buf) {
    for (var ii = 0; ii < 16; ++ii) {
      buf[i + ii] = rnds[ii];
    }
  }

  return buf || bytesToUuid_1(rnds);
}

var v4_1 = v4;

var ScannerContext = React.createContext({
  elements: {}
});

var checkVisible = function checkVisible(element) {
  var threshold = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 0;
  var mode = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : 'visible';

  var rect = element.getBoundingClientRect();
  var viewHeight = Math.max(document.documentElement.clientHeight, window.innerHeight);
  var above = rect.bottom - threshold < 0;
  var below = rect.top - viewHeight + threshold >= 0;

  var isVisible = !above && !below;
  if (mode === 'above') {
    isVisible = above;
  }

  if (mode === 'below') {
    isVisible = below;
  }

  return {
    isVisible: isVisible,
    above: above,
    below: below,
    rect: rect,
    viewHeight: viewHeight
  };
};

var checkVisibleAndScroll = function checkVisibleAndScroll(element) {
  var threshold = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 0;
  var mode = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : 'visible';

  var viewHeight = Math.max(document.documentElement.clientHeight, window.innerHeight);
  var fixedThreshold = viewHeight * threshold / 100;

  var _checkVisible = checkVisible(element, fixedThreshold, mode),
      isVisible = _checkVisible.isVisible,
      rect = _checkVisible.rect;

  var _window = window,
      scrollX = _window.scrollX,
      scrollY = _window.scrollY;


  if (!isVisible) {
    var elementPosY = scrollY + rect.top - fixedThreshold;
    window.scrollTo(scrollX, elementPosY);
    element.scrollIntoView();
  }
};

var classCallCheck$1 = function (instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
};

var createClass$1 = function () {
  function defineProperties(target, props) {
    for (var i = 0; i < props.length; i++) {
      var descriptor = props[i];
      descriptor.enumerable = descriptor.enumerable || false;
      descriptor.configurable = true;
      if ("value" in descriptor) descriptor.writable = true;
      Object.defineProperty(target, descriptor.key, descriptor);
    }
  }

  return function (Constructor, protoProps, staticProps) {
    if (protoProps) defineProperties(Constructor.prototype, protoProps);
    if (staticProps) defineProperties(Constructor, staticProps);
    return Constructor;
  };
}();

var _extends$2 = Object.assign || function (target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i];

    for (var key in source) {
      if (Object.prototype.hasOwnProperty.call(source, key)) {
        target[key] = source[key];
      }
    }
  }

  return target;
};

var inherits$1 = function (subClass, superClass) {
  if (typeof superClass !== "function" && superClass !== null) {
    throw new TypeError("Super expression must either be null or a function, not " + typeof superClass);
  }

  subClass.prototype = Object.create(superClass && superClass.prototype, {
    constructor: {
      value: subClass,
      enumerable: false,
      writable: true,
      configurable: true
    }
  });
  if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass;
};

var objectWithoutProperties$1 = function (obj, keys) {
  var target = {};

  for (var i in obj) {
    if (keys.indexOf(i) >= 0) continue;
    if (!Object.prototype.hasOwnProperty.call(obj, i)) continue;
    target[i] = obj[i];
  }

  return target;
};

var possibleConstructorReturn$1 = function (self, call) {
  if (!self) {
    throw new ReferenceError("this hasn't been initialised - super() hasn't been called");
  }

  return call && (typeof call === "object" || typeof call === "function") ? call : self;
};

var Scannable = function (_React$Component) {
  inherits$1(Scannable, _React$Component);

  function Scannable(props) {
    classCallCheck$1(this, Scannable);

    var _this = possibleConstructorReturn$1(this, (Scannable.__proto__ || Object.getPrototypeOf(Scannable)).call(this, props));

    _this.scannableId = v4_1();
    _this.isFocused = false;
    return _this;
  }

  createClass$1(Scannable, [{
    key: 'isEnabled',
    value: function isEnabled() {
      return !this.props.disabled;
    }
  }, {
    key: 'onSelect',
    value: function onSelect(event) {
      this.props.onSelect(event, this, this.props.scanner);
    }
  }, {
    key: 'componentDidMount',
    value: function componentDidMount() {
      var scanner = this.props.scanner;

      if (scanner && scanner.addScannableElement) {
        scanner.addScannableElement(this);
      }
    }
  }, {
    key: 'componentDidUpdate',
    value: function componentDidUpdate(_ref) {
      var disabled = _ref.disabled;

      if (disabled !== this.props.disabled) {
        var scanner = this.props.scanner;

        scanner.updateScannableElement(this);
      }
    }
  }, {
    key: 'render',
    value: function render() {
      var _this2 = this;

      var _props = this.props,
          children = _props.children,
          scanner = _props.scanner,
          disabled = _props.disabled,
          onFocus = _props.onFocus,
          onBlur = _props.onBlur,
          other = objectWithoutProperties$1(_props, ['children', 'scanner', 'disabled', 'onFocus', 'onBlur']);
      var focusedItem = scanner.focusedItem,
          config = scanner.config;


      var childrenWithProps = children;

      if (config) {
        var focusedClassName = config.focusedClassName,
            focusedVisibleThreshold = config.focusedVisibleThreshold;

        childrenWithProps = React.Children.map(children, function (child) {
          var classes = [child.props.className || ''];
          var isFocused = focusedItem.element && focusedItem.element.scannableId === _this2.scannableId;
          var isBlurred = _this2.isFocused && !isFocused;
          _this2.isFocused = isFocused;

          if (isFocused) {
            classes.push(focusedClassName);
            checkVisibleAndScroll(focusedItem.node, focusedVisibleThreshold);
            onFocus(_this2, scanner);
          } else if (isBlurred) {
            onBlur(_this2, scanner);
          }

          var className = classes.join(' ').trim();
          return React.cloneElement(child, _extends$2({ className: className }, other));
        });
      }

      return React.createElement(
        React.Fragment,
        null,
        childrenWithProps
      );
    }
  }]);
  return Scannable;
}(React.Component);

Scannable.defaultProps = {
  disabled: false,
  onFocus: function onFocus() {},
  onBlur: function onBlur() {},
  onSelect: function onSelect() {}
};

Scannable.propTypes = {
  children: PropTypes.node.isRequired,
  scanner: PropTypes.object.isRequired,
  disabled: PropTypes.bool,
  onFocus: PropTypes.func,
  onBlur: PropTypes.func,
  onSelect: PropTypes.func
};

var WithScanner = function WithScanner(Component$$1) {
  return function (props) {
    return React.createElement(
      ScannerContext.Consumer,
      null,
      function (scannerContext) {
        return React.createElement(Component$$1, _extends$2({}, props, { scanner: scannerContext }));
      }
    );
  };
};

var Scannable$1 = WithScanner(Scannable);

var SCANNER_STRATEGY = 'automatic';
var SCANNER_EVENTS = ['click', 'keydown', 'contextmenu'];
var SCANNER_CLASSNAME = 'Scanner__Container';
var SCANNER_SELECT_DEBOUNCE_TIME = 400;
var SCANNER_SELECT_KEYCODES = ['enter', 'backspace', 13, 8];
var SCANNER_AUTODEACTIVATE_KEYCODES = ['scape', 27];
var SCANNER_AUTODEACTIVATE_COUNT = 4;
var SCANNER_ADVANCE_KEYCODES = ['*', 'spacebar', 'tab', 'rightArrow', 'downArrow', 32, 9];
var SCANNER_SELECT_CLICKEVENT = 'contextmenu';
var SCANNER_ADVANCE_CLICKEVENT = 'click';
var SCANNER_MOVE_BACK_KEYS = ['leftArrow', 'upArrow'];
var SCANNER_CLASSNAME_ACTIVE = 'Scanner__Container Scanner__Container--active';
var SCANNER_ITERATION_INTERVAL = 2000;
var SCANNABLE_FOCUSED_CLASSNAME = 'scanner__focused';
var SCANNABLE_FOCUSED_VISIBLE_THRESHOLD = 5;

var KEY_CODE_MAP = {
  enter: 13,
  backspace: 8,
  spacebar: 32,
  tab: 9,
  shift: 16,
  control: 17,
  alt: 18,
  leftArrow: 37,
  upArrow: 38,
  rightArrow: 39,
  downArrow: 40,
  escape: 27
};

// https://gist.github.com/nmsdvid/8807205#gistcomment-2595895
// const debounce = (callback, time = 250, debounceInterval) => (...args) =>
//   clearTimeout(debounceInterval, (debounceInterval = setTimeout(() => callback(...args), time)));

var debounce$3 = function debounce(callback) {
  var time = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 250;
  var immediate = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : true;

  var timeout = void 0;
  return function () {
    for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    if (immediate && !timeout) {
      callback.apply(undefined, args);
    }

    clearTimeout(timeout);
    timeout = setTimeout(function () {
      timeout = null;
      if (!immediate) {
        callback.apply(undefined, args);
      }
    }, time);
  };
};

var dispatchEvent = function dispatchEvent(element, event) {
  var node = element.node;

  var eventToDispatch = new MouseEvent('click', {
    view: window,
    bubbles: true,
    cancelable: false
  });

  node.dispatchEvent(eventToDispatch);
};

var configKeys = ['advanceClickEvent', 'advanceKeyCodes', 'moveBackKeyCodes', 'autoDeactivateCount', 'autoDeactivateKeyCodes', 'className', 'classNameActive', 'events', 'iterationInterval', 'focusedClassName', 'focusedVisibleThreshold', 'selectClickEvent', 'selectDebounceTime', 'selectKeyCodes', 'strategy', 'target'];

var getConfig = function getConfig(props) {
  var config = {};
  var propsKeys = Object.keys(props).filter(function (k) {
    return configKeys.indexOf(k) >= 0;
  });

  propsKeys.forEach(function (k) {
    config[k] = props[k];
  });

  return config;
};

var BaseStrategy = function () {
  function BaseStrategy(scanner) {
    classCallCheck$1(this, BaseStrategy);

    this.scanner = scanner;
    this.config = scanner.config;
  }

  createClass$1(BaseStrategy, [{
    key: 'activate',
    value: function activate() {
      console.log('Scanner should be activated');
    }
  }, {
    key: 'deactivate',
    value: function deactivate() {
      console.log('Scanner should be deactivated');
    }
  }, {
    key: 'checkAutoDeactivation',
    value: function checkAutoDeactivation(event) {
      console.log('Strategy should provide a checkAutoDeactivation function');
    }
  }, {
    key: 'getNextScannableId',
    value: function getNextScannableId() {
      var focusedId = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : this.scanner.state.focusedId;
      var keysToIterate = this.scanner.state.keysToIterate;

      var nextFocusedIndex = keysToIterate.indexOf(focusedId) + 1;

      return nextFocusedIndex < keysToIterate.length ? keysToIterate[nextFocusedIndex] : keysToIterate[0];
    }
  }, {
    key: 'getPrevScannableId',
    value: function getPrevScannableId() {
      var focusedId = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : this.scanner.state.focusedId;
      var keysToIterate = this.scanner.state.keysToIterate;

      var nextFocusedIndex = keysToIterate.indexOf(focusedId) - 1;

      return nextFocusedIndex >= 0 ? keysToIterate[nextFocusedIndex] : keysToIterate[keysToIterate.length - 1];
    }
  }, {
    key: 'dispatchEvent',
    value: function dispatchEvent(scannable, event) {
      this.scanner.setSelectedElement(scannable, event);
      utils.dispatchEvent(scannable, event);
      scannable.element.onSelect(event);
    }
  }, {
    key: 'selectElement',
    value: function selectElement(scannable, event) {
      var selectedPath = [];
      if (scannable && scannable.treeId) {
        if (scannable.children) {
          selectedPath = this.scanner.state.selectedPath.concat(scannable.treeId);
        } else {
          this.dispatchEvent(scannable, event);
        }
      }

      this.scanner.setSelectedPath(selectedPath);
    }
  }]);
  return BaseStrategy;
}();

var AutomaticStrategy = function (_BaseStrategy) {
  inherits$1(AutomaticStrategy, _BaseStrategy);

  function AutomaticStrategy(scanner) {
    classCallCheck$1(this, AutomaticStrategy);

    var _this = possibleConstructorReturn$1(this, (AutomaticStrategy.__proto__ || Object.getPrototypeOf(AutomaticStrategy)).call(this, scanner));

    _this.autoDeactivationCounter = 0;
    _this.autoDeactivationToutFn = null;
    _this.autoDeactivationKeyCodes = new Set(scanner.config.autoDeactivateKeyCodes.map(function (kc) {
      return KEY_CODE_MAP[kc] || kc;
    }));
    return _this;
  }

  createClass$1(AutomaticStrategy, [{
    key: 'activate',
    value: function activate() {
      var _this2 = this;

      this.clearIterateInterval();

      if (this.scanner.state.keysToIterate.length) {
        this.iterateIntervalFn = setInterval(function () {
          var newFocusedId = _this2.getNextScannableId();
          _this2.scanner.focusScannable(newFocusedId);
        }, this.config.iterationInterval);
      }
    }
  }, {
    key: 'deactivate',
    value: function deactivate() {
      this.clearIterateInterval();
    }
  }, {
    key: 'checkAutoDeactivation',
    value: function checkAutoDeactivation(event) {
      var _this3 = this;

      var eventType = event.type,
          keyCode = event.keyCode;

      if (eventType === 'keydown') {
        if (this.autoDeactivationKeyCodes.has(keyCode)) {
          if (this.autoDeactivationToutFn) {
            clearTimeout(this.autoDeactivationToutFn);
          }

          this.autoDeactivationCounter = this.autoDeactivationCounter + 1;
          if (this.autoDeactivationCounter >= this.scanner.config.autoDeactivateCount) {
            this.deactivate();
            this.scanner.triggerDeactivation();
            this.autoDeactivationCounter = 0;
          } else {
            this.autoDeactivationToutFn = setTimeout(function () {
              _this3.autoDeactivationCounter = 0;
            }, 600);
          }

          return false;
        }
      }

      return true;
    }
  }, {
    key: 'clearIterateInterval',
    value: function clearIterateInterval() {
      if (this.iterateIntervalFn) {
        clearInterval(this.iterateIntervalFn);
      }
    }
  }]);
  return AutomaticStrategy;
}(BaseStrategy);

var ManualStrategy = function (_BaseStrategy) {
  inherits$1(ManualStrategy, _BaseStrategy);

  function ManualStrategy(scanner) {
    classCallCheck$1(this, ManualStrategy);

    var _this = possibleConstructorReturn$1(this, (ManualStrategy.__proto__ || Object.getPrototypeOf(ManualStrategy)).call(this, scanner));

    var _scanner$config = scanner.config,
        selectClickEvent = _scanner$config.selectClickEvent,
        advanceClickEvent = _scanner$config.advanceClickEvent,
        moveBackKeyCodes = _scanner$config.moveBackKeyCodes,
        selectKeyCodes = _scanner$config.selectKeyCodes,
        advanceKeyCodes = _scanner$config.advanceKeyCodes,
        autoDeactivateKeyCodes = _scanner$config.autoDeactivateKeyCodes;

    _this.selectClickEvent = selectClickEvent;
    _this.advanceClickEvent = advanceClickEvent;
    _this.selectKeyCodes = new Set(selectKeyCodes.map(function (kc) {
      return KEY_CODE_MAP[kc] || kc;
    }));
    _this.advanceKeyCodes = new Set(advanceKeyCodes.map(function (kc) {
      return KEY_CODE_MAP[kc] || kc;
    }));
    _this.moveBackKeyCodes = new Set(moveBackKeyCodes.map(function (kc) {
      return KEY_CODE_MAP[kc] || kc;
    }));

    _this.autoDeactivationCounter = 0;
    _this.autoDeactivationToutFn = null;
    _this.autoDeactivationKeyCodes = new Set(autoDeactivateKeyCodes.map(function (kc) {
      return KEY_CODE_MAP[kc] || kc;
    }));

    _this.isActive = false;
    return _this;
  }

  createClass$1(ManualStrategy, [{
    key: 'activate',
    value: function activate() {
      this.isActive = true;
    }
  }, {
    key: 'deactivate',
    value: function deactivate() {
      this.isActive = false;
    }
  }, {
    key: 'checkAutoDeactivation',
    value: function checkAutoDeactivation(event) {
      var _this2 = this;

      var eventType = event.type,
          keyCode = event.keyCode;

      if (eventType === 'keydown') {
        if (this.autoDeactivationKeyCodes.has(keyCode)) {
          if (this.autoDeactivationToutFn) {
            clearTimeout(this.autoDeactivationToutFn);
          }

          this.autoDeactivationCounter = this.autoDeactivationCounter + 1;
          if (this.autoDeactivationCounter >= this.scanner.config.autoDeactivateCount) {
            this.deactivate();
            this.scanner.triggerDeactivation();
            this.autoDeactivationCounter = 0;
          } else {
            this.autoDeactivationToutFn = setTimeout(function () {
              _this2.autoDeactivationCounter = 0;
            }, 600);
          }

          return false;
        }
      }

      return true;
    }
  }, {
    key: 'selectElement',
    value: function selectElement(scannable, event) {
      var eventType = event.type,
          keyCode = event.keyCode;

      var reverse = this.moveBackKeyCodes.has(keyCode);

      switch (eventType) {
        case 'keydown':
          if (this.selectKeyCodes.has(keyCode)) {
            this.select(scannable, event);
          } else if (reverse || this.advanceKeyCodes.has(keyCode) || this.advanceKeyCodes.has('*')) {
            this.advance(scannable, event, reverse);
          }
          break;

        case this.selectClickEvent:
          this.select(scannable, event);
          break;

        case this.advanceClickEvent:
          this.advance(scannable, event);
          break;

        default:
          break;
      }
    }
  }, {
    key: 'advance',
    value: function advance(scannable, event) {
      var reverse = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : false;

      var action = reverse ? 'getPrevScannableId' : 'getNextScannableId';
      var newFocusedId = this[action]();
      this.scanner.focusScannable(newFocusedId);
    }
  }, {
    key: 'select',
    value: function select(scannable, event) {
      var selectedPath = [];
      if (scannable && scannable.treeId) {
        if (scannable.children) {
          selectedPath = this.scanner.state.selectedPath.concat(scannable.treeId);
        } else {
          this.dispatchEvent(scannable, event);
        }
      }

      this.scanner.setSelectedPath(selectedPath);
    }
  }]);
  return ManualStrategy;
}(BaseStrategy);

var Strategies = {
  automatic: AutomaticStrategy,
  manual: ManualStrategy
};

var getStrategy = function getStrategy(strategyName, scanner) {
  var StrategyConstructor = Strategies[strategyName] || Strategies.automatic;
  return new StrategyConstructor(scanner);
};

var getScannableChildrenOfElement = function getScannableChildrenOfElement(element, scannables) {
  var scannableNodes = scannables.map(function (_ref) {
    var node = _ref.node;
    return node;
  });
  var elementChildren = Array.from(element.children || []);
  var scannableChildren = [];

  if (elementChildren.length) {
    elementChildren.forEach(function (child) {
      var index = scannableNodes.indexOf(child);
      if (index >= 0) {
        var scannableChild = scannables[index];
        scannableChildren.push(scannableChild);
      } else if (child.children && child.children.length) {
        scannableChildren = scannableChildren.concat(getScannableChildrenOfElement(child, scannables));
      }
    });
  }

  return scannableChildren;
};

var getTreeForElement = function getTreeForElement(elementNode, scannables) {
  var scannableChildren = getScannableChildrenOfElement(elementNode, scannables);

  var newScannables = scannables.filter(function (e) {
    return scannableChildren.indexOf(e) < 0;
  });
  scannableChildren.forEach(function (sc) {
    if (sc.node.children && sc.node.children.length) {
      sc.children = getTreeForElement(sc.node, newScannables);
    }
  });

  var tree = scannableChildren.reduce(function (prev, sc, i) {
    var treeId = i + "_" + sc.element.scannableId;
    prev[treeId] = _extends$2({}, sc, { treeId: treeId });

    if (sc.children && Object.keys(sc.children).length === 0) {
      delete prev[treeId].children;
    }

    return prev;
  }, {});

  return tree;
};

var utils = {
  checkVisible: checkVisible,
  checkVisibleAndScroll: checkVisibleAndScroll,
  debounce: debounce$3,
  dispatchEvent: dispatchEvent,
  getConfig: getConfig,
  getStrategy: getStrategy,
  getTreeForElement: getTreeForElement
};

function styleInject$1(css, ref) {
  if ( ref === void 0 ) ref = {};
  var insertAt = ref.insertAt;

  if (!css || typeof document === 'undefined') { return; }

  var head = document.head || document.getElementsByTagName('head')[0];
  var style = document.createElement('style');
  style.type = 'text/css';

  if (insertAt === 'top') {
    if (head.firstChild) {
      head.insertBefore(style, head.firstChild);
    } else {
      head.appendChild(style);
    }
  } else {
    head.appendChild(style);
  }

  if (style.styleSheet) {
    style.styleSheet.cssText = css;
  } else {
    style.appendChild(document.createTextNode(css));
  }
}

var css$9 = ".Scanner__Container {\n  position: relative;\n  width: 100%;\n  height: 100%;\n  overflow: auto;\n}\n\n.Scanner__EventsHandler {\n  position: absolute;\n  left: 0;\n  top: 0;\n  right: 0;\n  bottom: 0;\n  background: rgba(0, 0, 0, 0.3);\n}\n\n.scanner__focused {\n  outline: 4px solid rgba(255, 0, 0, 1);\n  outline-offset: -4px;\n}\n";
styleInject$1(css$9);

var getConfig$1 = utils.getConfig,
    getStrategy$1 = utils.getStrategy,
    getTreeForElement$1 = utils.getTreeForElement,
    debounce$1$1 = utils.debounce;


var SCANNER_INITIAL_STATE = {
  selectedPath: [],
  elementsToIterate: {},
  keysToIterate: [],
  focusedId: 'not-valid-id'
};

var Scanner = function (_React$Component) {
  inherits$1(Scanner, _React$Component);

  function Scanner(props) {
    classCallCheck$1(this, Scanner);

    var _this = possibleConstructorReturn$1(this, (Scanner.__proto__ || Object.getPrototypeOf(Scanner)).call(this, props));

    _this.scannerEventAction = function (event) {
      var active = _this.props.active;


      var shouldContinue = true;
      if (_this.config.autoDeactivateCount) {
        shouldContinue = _this.strategy.checkAutoDeactivation(event);
      }

      if (!shouldContinue) {
        _this.props.onAutoDeactivateAction(event);
      }

      if (active && shouldContinue && !_this.selectedElement) {
        event.preventDefault();
        event.stopPropagation();

        var elementToSelect = _this.state.elementsToIterate[_this.state.focusedId];
        if (elementToSelect) {
          _this.debouncedSelectElement(elementToSelect, event);
        }
      }
    };

    _this.addScannableElement = function (element) {
      _this.elements[element.scannableId] = {
        element: element,
        node: ReactDOM.findDOMNode(element)
      };

      if (_this.props.active) {
        _this.findNodes();
      }
    };

    _this.updateScannableElement = function (element) {
      _this.elements[element.scannableId] = {
        element: element,
        node: ReactDOM.findDOMNode(element)
      };

      var focusedId = _this.state.focusedId;
      if (element.scannableId === focusedId) {
        focusedId = _this.strategy.getNextScannableId(focusedId);
      }

      _this.findNodes();
      _this.iterateScannableElements(focusedId);
    };

    _this.onScannableFocus = function () {
      var focusedId = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : _this.state.focusedId;

      var _ref = _this.state.elementsToIterate[focusedId] || {},
          element = _ref.element;

      if (element) {
        _this.props.onFocus(element, _this);
      }
    };

    _this.reset = function () {
      _this.strategy.deactivate();
      _this.findNodes();
      _this.setState(SCANNER_INITIAL_STATE, _this.iterateScannableElements);
    };

    _this.state = SCANNER_INITIAL_STATE;
    _this.scannerNode = null;
    _this.elements = {};
    _this.tree = {};
    _this.selectedElement = null;
    _this.config = getConfig$1(props);
    _this.strategy = getStrategy$1(_this.config.strategy, _this);
    _this.debouncedSelectElement = debounce$1$1(function (element, event) {
      _this.selectElement(element, event);
    }, _this.config.selectDebounceTime);

    _this.target = _this.config.target;
    return _this;
  }

  createClass$1(Scanner, [{
    key: 'componentDidMount',
    value: function componentDidMount() {
      this.registerEvents();

      if (this.props.active) {
        this.findNodes();
        this.iterateScannableElements();
      }
    }
  }, {
    key: 'componentWillUnmount',
    value: function componentWillUnmount() {
      this.unregisterEvents();
      this.strategy.deactivate();
    }
  }, {
    key: 'componentDidUpdate',
    value: function componentDidUpdate(prevProps) {
      var _this2 = this;

      if (prevProps.active !== this.props.active) {
        if (this.props.active) {
          this.findNodes();
          this.iterateScannableElements();
        } else {
          this.strategy.deactivate();
          this.setState(SCANNER_INITIAL_STATE);
        }
      }

      var newConfig = getConfig$1(this.props);
      if (newConfig.strategy !== this.config.strategy) {
        this.strategy.deactivate();
        this.strategy = getStrategy$1(newConfig.strategy, this);
      }

      this.config = newConfig;
      this.debouncedSelectElement = debounce$1$1(function (element, event) {
        _this2.selectElement(element, event);
      }, this.config.selectDebounceTime);
    }
  }, {
    key: 'triggerDeactivation',
    value: function triggerDeactivation() {
      this.props.onDeactivation();
      this.setState(SCANNER_INITIAL_STATE);
    }
  }, {
    key: 'findNodes',
    value: function findNodes() {
      this.scannerNode = ReactDOM.findDOMNode(this);
      var elementsArray = Object.values(this.elements);
      var children = getTreeForElement$1(this.scannerNode, elementsArray);
      this.tree = { children: children };
    }
  }, {
    key: 'getElementsToIterate',
    value: function getElementsToIterate() {
      var active = this.props.active;

      var elementsToIterate = {};
      var keysToIterate = [];
      if (!active) {
        return {
          elementsToIterate: elementsToIterate,
          keysToIterate: keysToIterate
        };
      }

      var current = this.tree;
      if (this.state.selectedPath.length) {
        current = this.state.selectedPath.reduce(function (prev, id) {
          prev = prev.children[id];
          return prev;
        }, current);
      }

      if (!current || !current.children) {
        current = {};
      }

      elementsToIterate = current.children || {};
      keysToIterate = Object.keys(elementsToIterate).filter(function (k) {
        return elementsToIterate[k].element.isEnabled();
      });

      return {
        elementsToIterate: elementsToIterate,
        keysToIterate: keysToIterate
      };
    }
  }, {
    key: 'selectElement',
    value: function selectElement(scannable, event) {
      if (scannable && scannable.treeId) {
        this.strategy.selectElement(scannable, event);
      }
    }
  }, {
    key: 'setSelectedElement',
    value: function setSelectedElement(scannable, event) {
      this.selectedElement = scannable;

      this.props.onSelect(event, scannable, this);
    }
  }, {
    key: 'setSelectedPath',
    value: function setSelectedPath() {
      var selectedPath = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : this.state.selectedPath;

      this.setState({ selectedPath: selectedPath }, this.iterateScannableElements);
    }
  }, {
    key: 'focusScannable',
    value: function focusScannable(focusedId) {
      this.setState({ focusedId: focusedId }, this.onScannableFocus);
    }
  }, {
    key: 'iterateScannableElements',
    value: function iterateScannableElements() {
      var _this3 = this;

      var focusedId = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : this.state.focusedId;

      this.selectedElement = null;

      var _getElementsToIterate = this.getElementsToIterate(),
          elementsToIterate = _getElementsToIterate.elementsToIterate,
          keysToIterate = _getElementsToIterate.keysToIterate;

      var newFocusedId = elementsToIterate[focusedId] ? focusedId : keysToIterate[0];

      this.setState({
        elementsToIterate: elementsToIterate,
        keysToIterate: keysToIterate,
        focusedId: newFocusedId
      }, function () {
        _this3.onScannableFocus(newFocusedId);
        _this3.strategy.activate();
      });
    }
  }, {
    key: 'registerEvents',
    value: function registerEvents() {
      var _this4 = this;

      this.config.events.forEach(function (e) {
        return _this4.target.addEventListener(e, _this4.scannerEventAction);
      });
    }
  }, {
    key: 'unregisterEvents',
    value: function unregisterEvents() {
      var _this5 = this;

      this.config.events.forEach(function (e) {
        return _this5.target.removeEventListener(e, _this5.scannerEventAction);
      });
    }
  }, {
    key: 'render',
    value: function render() {
      var _props = this.props,
          children = _props.children,
          active = _props.active;

      var focusedItem = this.state.keysToIterate.length > 0 ? this.state.elementsToIterate[this.state.focusedId] : {};

      var contextValue = _extends$2({}, this.state, {
        config: this.config,
        focusedItem: focusedItem,
        addScannableElement: this.addScannableElement,
        updateScannableElement: this.updateScannableElement,
        reset: this.reset
      });

      var classes = active ? this.config.classNameActive : this.config.className;

      return React.createElement(
        ScannerContext.Provider,
        { value: contextValue },
        React.createElement(
          'div',
          { className: classes },
          children
        )
      );
    }
  }]);
  return Scanner;
}(React.Component);

Scanner.defaultProps = {
  active: false,
  autoDeactivateCount: SCANNER_AUTODEACTIVATE_COUNT,
  autoDeactivateKeyCodes: SCANNER_AUTODEACTIVATE_KEYCODES,
  advanceKeyCodes: SCANNER_ADVANCE_KEYCODES,
  advanceClickEvent: SCANNER_ADVANCE_CLICKEVENT,
  moveBackKeyCodes: SCANNER_MOVE_BACK_KEYS,
  className: SCANNER_CLASSNAME,
  classNameActive: SCANNER_CLASSNAME_ACTIVE,
  events: SCANNER_EVENTS,
  focusedClassName: SCANNABLE_FOCUSED_CLASSNAME,
  focusedVisibleThreshold: SCANNABLE_FOCUSED_VISIBLE_THRESHOLD,
  iterationInterval: SCANNER_ITERATION_INTERVAL,
  onSelect: function onSelect() {},
  onFocus: function onFocus() {},
  onAutoDeactivateAction: function onAutoDeactivateAction() {},
  onDeactivation: function onDeactivation() {},
  selectClickEvent: SCANNER_SELECT_CLICKEVENT,
  selectDebounceTime: SCANNER_SELECT_DEBOUNCE_TIME,
  selectKeyCodes: SCANNER_SELECT_KEYCODES,
  strategy: SCANNER_STRATEGY,
  target: document.body
};

Scanner.propTypes = {
  active: PropTypes.bool,
  autoDeactivateCount: PropTypes.number,
  autoDeactivateKeyCodes: PropTypes.arrayOf(PropTypes.oneOfType([PropTypes.string, PropTypes.number])),
  advanceKeyCodes: PropTypes.arrayOf(PropTypes.oneOfType([PropTypes.string, PropTypes.number])),
  advanceClickEvent: PropTypes.string,
  moveBackKeyCodes: PropTypes.arrayOf(PropTypes.oneOfType([PropTypes.string, PropTypes.number])),
  children: PropTypes.node.isRequired,
  className: PropTypes.string,
  classNameActive: PropTypes.string,
  events: PropTypes.arrayOf(PropTypes.string),
  focusedClassName: PropTypes.string,
  focusedVisibleThreshold: PropTypes.number,
  iterationInterval: PropTypes.number,
  onSelect: PropTypes.func,
  onFocus: PropTypes.func,
  onAutoDeactivateAction: PropTypes.func,
  onDeactivation: PropTypes.func,
  selectClickEvent: PropTypes.string,
  selectDebounceTime: PropTypes.number,
  selectKeyCodes: PropTypes.arrayOf(PropTypes.oneOfType([PropTypes.string, PropTypes.number])),
  strategy: PropTypes.string,
  target: PropTypes.instanceOf(Element)
};

var getImageSrc = function getImageSrc(image) {
  return image ? image.data ||
  /* image.path || */
  image.url : '';
};

var getSoundSrc = function getSoundSrc(sound) {
  return sound ? sound.data || sound.path || sound.url : '';
};

var getLocaleMessage = function getLocaleMessage(board, message) {
  var locale = board.locale,
      strings = board.strings;
  var localeMessage = strings && strings[locale] && strings[locale][message];
  return localeMessage || message;
};

var denormalizeBoardButtons = function denormalizeBoardButtons(board) {
  return board.buttons.map(function (_ref) {
    var image_id = _ref.image_id,
        label = _ref.label,
        sound_id = _ref.sound_id,
        vocalization = _ref.vocalization,
        other = _objectWithoutProperties(_ref, ["image_id", "label", "sound_id", "vocalization"]);

    return _objectSpread({}, other, {
      image: getImageSrc(board.images.find(function (image) {
        return image.id === image_id;
      })),
      label: getLocaleMessage(board, label),
      sound: getSoundSrc(board.sounds.find(function (sound) {
        return sound.id === sound_id;
      })),
      vocalization: getLocaleMessage(board, vocalization)
    });
  });
};
/**
 * Board component, render OBF compliant objects
 */


var Board =
/*#__PURE__*/
function (_Component) {
  _inherits(Board, _Component);

  function Board() {
    var _getPrototypeOf2;

    var _this;

    _classCallCheck(this, Board);

    for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    _this = _possibleConstructorReturn(this, (_getPrototypeOf2 = _getPrototypeOf(Board)).call.apply(_getPrototypeOf2, [this].concat(args)));

    _defineProperty(_assertThisInitialized(_assertThisInitialized(_this)), "state", {
      buttons: [],
      output: []
    });

    _defineProperty(_assertThisInitialized(_assertThisInitialized(_this)), "changeOutput", function (output) {
      var onOutputChange = _this.props.onOutputChange;

      if (onOutputChange) {
        onOutputChange(output);
      } else {
        _this.setState({
          output: output
        });
      }
    });

    _defineProperty(_assertThisInitialized(_assertThisInitialized(_this)), "handleBoardButtonClick", function (button) {
      var _this$props = _this.props,
          onLoadBoard = _this$props.onLoadBoard,
          onBoardButtonClick = _this$props.onBoardButtonClick,
          onPlaySound = _this$props.onPlaySound,
          onSpeak = _this$props.onSpeak;

      if (button.vocalization) {
        onSpeak && onSpeak(button.vocalization);
      }

      if (button.actions && button.actions.length) {
        button.actions.forEach(function (action) {
          return _this.doAction(action);
        });
        return;
      } else if (button.action) {
        _this.doAction(button.action);

        return;
      }

      if (button.load_board) {
        onLoadBoard && onLoadBoard(button.load_board);
        return;
      }

      if (button.sound) {
        onPlaySound && onPlaySound(button.sound);
      } else if (button.label) {
        onSpeak && onSpeak(button.label);
      }

      var symbol = {
        label: button.label,
        sound: button.sound,
        image: button.image,
        vocalization: button.vocalization
      };

      _this.outputPush(symbol);

      onBoardButtonClick && onBoardButtonClick(button);
    });

    _defineProperty(_assertThisInitialized(_assertThisInitialized(_this)), "handleOutputClick", function () {
      var onSpeak = _this.props.onSpeak;

      var text = _this.reduceOutput();

      onSpeak && onSpeak(text);
    });

    _defineProperty(_assertThisInitialized(_assertThisInitialized(_this)), "renderButton", function (button) {
      var renderButton = _this.props.renderButton;
      return button.hidden ? null : React.createElement(Scannable$1, null, React.cloneElement(renderButton(button), {
        onClick: function onClick() {
          _this.handleBoardButtonClick(button);
        }
      }));
    });

    _defineProperty(_assertThisInitialized(_assertThisInitialized(_this)), "renderSymbol", function (_ref2) {
      var label = _ref2.label,
          image = _ref2.image;
      return React.createElement(_Symbol, {
        label: label,
        src: image
      });
    });

    _defineProperty(_assertThisInitialized(_assertThisInitialized(_this)), "renderOutput", function () {
      var _this$props2 = _this.props,
          backspaceButton = _this$props2.backspaceButton,
          clearButton = _this$props2.clearButton,
          dir = _this$props2.dir,
          outputHidden = _this$props2.outputHidden;
      return outputHidden ? null : React.createElement(Scannable$1, {
        disabled: !_this.state.output.length
      }, React.createElement(Output, {
        backspaceButton: backspaceButton,
        clearButton: clearButton,
        dir: dir,
        interactiveWrapper: Scannable$1,
        onChange: _this.changeOutput,
        onClick: _this.handleOutputClick,
        renderValue: _this.renderSymbol,
        values: _this.state.output
      }));
    });

    _defineProperty(_assertThisInitialized(_assertThisInitialized(_this)), "renderNavbar", function () {
      var _this$props3 = _this.props,
          board = _this$props3.board,
          navbar = _this$props3.navbar;
      return navbar ? navbar : React.createElement("div", {
        className: "Board__name"
      }, React.createElement("div", null, board.name));
    });

    _defineProperty(_assertThisInitialized(_assertThisInitialized(_this)), "renderGrid", function () {
      var grid = _this.props.board.grid;
      return React.createElement(reactSizeme_1, {
        monitorHeight: true
      }, function (_ref3) {
        var size = _ref3.size;
        return React.createElement(Scannable$1, null, React.createElement(Grid, {
          columns: grid.columns,
          items: _this.state.buttons,
          order: grid.order,
          renderItem: _this.renderButton,
          rows: grid.rows,
          rowWrapper: Scannable$1,
          size: size
        }));
      });
    });

    return _this;
  }

  _createClass(Board, [{
    key: "outputClear",
    value: function outputClear() {
      var output = [];
      this.changeOutput(output);
    }
  }, {
    key: "outputPush",
    value: function outputPush(symbol) {
      var output = _toConsumableArray(this.state.output).concat([symbol]);

      this.changeOutput(output);
    }
  }, {
    key: "outputJoinLabel",
    value: function outputJoinLabel(label) {
      if (this.state.output.length === 0) {
        return;
      }

      var output = _toConsumableArray(this.state.output);

      var lastSymbol = output[output.length - 1];
      lastSymbol.label = "".concat(lastSymbol.label).concat(label);
      this.changeOutput(output);
    }
  }, {
    key: "doAction",
    value: function doAction(action) {
      switch (action) {
        case ButtonActions.clear:
          this.outputClear();
          break;

        default: // no default

      }

      if (action.startsWith('+')) {
        var label = action.slice(1);
        this.outputJoinLabel(label);
      }
    }
  }, {
    key: "reduceOutput",
    value: function reduceOutput() {
      return this.state.output.reduce(function (previousText, currentValue) {
        var currentText = currentValue.vocalization || currentValue.label;
        return "".concat(previousText, " ").concat(currentText);
      }, '');
    }
  }, {
    key: "render",
    value: function render() {
      var _this$props4 = this.props,
          dir = _this$props4.dir,
          scanInterval = _this$props4.scanInterval,
          scanning = _this$props4.scanning,
          size = _this$props4.size,
          toolbar = _this$props4.toolbar;
      var output = this.renderOutput();
      var navbar = this.renderNavbar();
      var grid = this.renderGrid();
      return React.createElement(Scanner, {
        active: scanning,
        iterationInterval: scanInterval
      }, React.createElement(BoardLayout, {
        dir: dir,
        size: size,
        output: output,
        navbar: navbar,
        toolbar: toolbar,
        grid: grid
      }));
    }
  }], [{
    key: "getDerivedStateFromProps",
    value: function getDerivedStateFromProps(props, state) {
      // TODO optimize getDerivedStateFromProps
      var buttons = denormalizeBoardButtons(props.board);
      var output = props.output || state.output;
      return _objectSpread({}, state, {
        buttons: buttons,
        output: output
      });
    }
  }]);

  return Board;
}(Component);

_defineProperty(Board, "propTypes", {
  /**
   * Component to render backspace button.
   */
  backspaceButton: PropTypes.node,

  /**
   * Board to display, needs to be [Open Board Format](http://www.openboardformat.org/) compliant.
   */
  board: PropTypes.shape({
    /**
     * ID
     */
    id: PropTypes.string,

    /**
     * Locale
     */
    locale: PropTypes.string,

    /**
     * Name
     */
    name: PropTypes.string,

    /**
     * Buttons
     */
    buttons: PropTypes.arrayOf(PropTypes.shape({
      action: PropTypes.string,
      id: PropTypes.oneOfType([PropTypes.string, PropTypes.number]).isRequired,
      label: PropTypes.string,
      vocalization: PropTypes.string,
      image_id: PropTypes.string,
      sound_id: PropTypes.string,
      background_color: PropTypes.string,
      border_color: PropTypes.string,
      load_board: PropTypes.shape({
        name: PropTypes.string,
        url: PropTypes.string,
        data_url: PropTypes.string
      }),
      hidden: PropTypes.bool
    })),

    /**
     * Images
     */
    images: PropTypes.arrayOf(PropTypes.shape({
      id: PropTypes.oneOfType([PropTypes.string, PropTypes.number]).isRequired,
      width: PropTypes.number,
      height: PropTypes.number,
      data: PropTypes.string,
      'content-type': PropTypes.string,
      url: PropTypes.string,
      source_url: PropTypes.string,
      author_name: PropTypes.string,
      author_url: PropTypes.string
    })),

    /**
     * Sounds
     */
    sounds: PropTypes.arrayOf(PropTypes.shape({})),

    /**
     * Grid
     */
    grid: PropTypes.shape({
      /**
       * Number of rows
       */
      rows: PropTypes.number,

      /**
       * Number of columns
       */
      columns: PropTypes.number,

      /**
       * Buttons order
       */
      order: PropTypes.arrayOf(PropTypes.arrayOf(PropTypes.oneOfType([PropTypes.string, PropTypes.number])))
    })
  }),

  /**
   * Component to render clear button.
   */
  clearButton: PropTypes.node,

  /**
   * Text direction
   */
  dir: PropTypes.oneOf(['ltr', 'rtl']),

  /**
   * Component to render navigation bar.
   */
  navbar: PropTypes.node,

  /**
   * Callback, fired when requesting to load board.
   * @param {Object} board
   */
  onLoadBoard: PropTypes.func,

  /**
   * Callback, fired when output changes.
   * @param {Array} output
   */
  onOutputChange: PropTypes.func,

  /**
   * Callback, fired when requesting to play sound.
   * @param {Object} sound - Sound to play.
   */
  onPlaySound: PropTypes.func,

  /**
   * Callback, fired when requesting to speak.
   * @param {string} text - Text to speak.
   */
  onSpeak: PropTypes.func,

  /**
   * Output to display.
   */
  output: PropTypes.arrayOf(PropTypes.shape({
    /**
     * Image src to display.
     */
    image: PropTypes.string,

    /**
     * Label to display.
     */
    label: PropTypes.string,

    /**
     * Sound to play.
     */
    sound: PropTypes.string,

    /**
     * Overrides `label` property in speech output.
     */
    vocalization: PropTypes.string
  })),

  /**
   * When `true`, output is hidden
   */
  outputHidden: PropTypes.bool,

  /**
   * Button renderer
   * @param {Object}
   */
  renderButton: PropTypes.func,

  /**
   * Sets the scan rate interval when `scanning` is `true`.
   */
  scanInterval: PropTypes.number,

  /**
   * When `true`, scanner is activated.
   */
  scanning: PropTypes.bool,

  /**
   * UI size.
   */
  size: PropTypes.oneOf(['default', 'large', 'larger']),

  /**
   * Component to render toolbar.
   */
  toolbar: PropTypes.node
});

_defineProperty(Board, "defaultProps", {
  dir: 'ltr',
  scanInterval: 1500,
  size: 'default'
});

var css$a = ".Tile {\r\n  position: relative;\r\n  box-sizing: border-box;\r\n  height: 100%;\r\n  width: 100%;\r\n  padding: 4px;\r\n  border: 2px solid transparent;\r\n  overflow: hidden;\r\n}\r\n\r\n/* overlay effect */\r\n.Tile::after {\r\n  content: '';\r\n  position: absolute;\r\n  top: 0;\r\n  right: 0;\r\n  bottom: 0;\r\n  left: 0;\r\n  background: rgba(0, 0, 0, 0.25);\r\n  opacity: 0;\r\n}\r\n\r\n.Tile:hover::after,\r\n.Tile:focus::after {\r\n  opacity: 0.8;\r\n}\r\n\r\n.Tile:active::after {\r\n  opacity: 1;\r\n}\r\n\r\n.Tile--folder::before {\r\n  content: '';\r\n  position: absolute;\r\n  top: 0;\r\n  right: 0;\r\n  width: 0;\r\n  height: 0;\r\n  border-top: 16px solid transparent;\r\n  border-bottom: 16px solid transparent;\r\n  border-left: 16px solid #000;\r\n  transform-origin: center right;\r\n  transform: translateY(-16px) rotate(-45deg);\r\n  z-index: 1;\r\n}\r\n";
styleInject(css$a);

var WcagContrastRatio = {
  AAA: 7,
  AA: 4.5
};

function getContrastColor(baseColor) {
  var lightColor = 'rgb(255, 255, 255)';
  var darkColor = 'rgb(0, 0, 0)';
  var contrast = colorContrast(baseColor, darkColor);
  var color = contrast < WcagContrastRatio.AA ? lightColor : darkColor;
  return color;
}

var propTypes$3 = {
  /**
   * Background color
   */
  background_color: PropTypes.string,

  /**
   * Border color
   */
  border_color: PropTypes.string,

  /**
   * Tile content
   */
  children: PropTypes.node,

  /**
   * @ignore
   */
  className: PropTypes.string,

  /**
   * Image src to display
   */
  image: PropTypes.string,

  /**
   * Label to display
   */
  label: PropTypes.string,

  /**
   * Board to load
   */
  load_board: PropTypes.shape({})
};
var defaultProps = {
  background_color: 'rgb(255, 255, 255)'
};

var Tile = function Tile(props) {
  var background_color = props.background_color,
      border_color = props.border_color,
      children = props.children,
      classNameProp = props.className,
      id = props.id,
      image = props.image,
      label = props.label,
      load_board = props.load_board,
      sound = props.sound,
      styleProp = props.style,
      other = _objectWithoutProperties(props, ["background_color", "border_color", "children", "className", "id", "image", "label", "load_board", "sound", "style"]);

  var folder = load_board && 'folder';
  var className = classnames('Tile', classNameProp, {
    'Tile--folder': folder
  });
  var contrastColor = getContrastColor(background_color);

  var style = _objectSpread({}, styleProp, {
    backgroundColor: background_color,
    borderColor: border_color,
    color: contrastColor
  });

  return React.createElement("button", _extends({
    type: "button",
    className: className,
    style: style
  }, other), React.createElement(_Symbol, {
    label: label,
    src: image
  }), children);
};

Tile.propTypes = propTypes$3;
Tile.defaultProps = defaultProps;

export { Board, Output, _Symbol as Symbol, Tile };
